__all__ = ['Planet']

import os
import numpy as np

from pylightcurve.errors import *
from pylightcurve.models.exoplanet_lc import planet_orbit, planet_star_projected_distance, planet_phase, \
    transit, transit_integrated, \
    transit_duration, transit_depth, eclipse, eclipse_integrated, eclipse_mid_time, eclipse_duration, eclipse_depth,\
    exotethys, fp_over_fs, _get_filter
from pylightcurve.analysis.optimisation import EmceeFitting
from pylightcurve.processes.files import open_dict, save_dict
from pylightcurve.plots.plots_fitting import plot_transit_fitting_models
from pylightcurve.spacetime.angles import Degrees, _request_angle
from pylightcurve.spacetime.targets import FixedTarget


class Filter:

    def __init__(self, rp_over_rs, ldc1, ldc2, ldc3, ldc4, fp_over_fs):

        self.rp_over_rs = rp_over_rs
        self.fp_over_fs = fp_over_fs
        self.limb_darkening_coefficients = [ldc1, ldc2, ldc3, ldc4]


class Planet:

    def __init__(self, name, ra, dec, stellar_logg, stellar_temperature, stellar_metallicity,
                 rp_over_rs, period, sma_over_rs, eccentricity, inclination,
                 periastron, mid_time, mid_time_format, ww=0,
                 albedo=0.15, emissivity=1.0,
                 ldc_method='claret', ldc_stellar_model='phoenix'):

        if isinstance(ra, float):
            ra = Degrees(ra)
        else:
            _request_angle(ra)

        if isinstance(dec, float):
            dec = Degrees(dec)
        else:
            _request_angle(dec)

        if isinstance(inclination, float):
            pass
        else:
            _request_angle(inclination)
            inclination = inclination.deg()

        if isinstance(periastron, float):
            pass
        else:
            _request_angle(periastron)
            periastron = periastron.deg()

        self.name = name
        self.target = FixedTarget(ra, dec)

        self.stellar_logg = stellar_logg
        self.stellar_temperature = stellar_temperature
        self.stellar_metallicity = stellar_metallicity
        self.ldc_method = ldc_method
        self.ldc_stellar_model = ldc_stellar_model

        self.rp_over_rs = rp_over_rs
        self.period = period
        self.sma_over_rs = sma_over_rs
        self.eccentricity = eccentricity
        self.inclination = inclination
        self.periastron = periastron
        self.mid_time = self.target.convert_to_bjd_tdb(mid_time, mid_time_format)
        self.eclipse_mid_time = eclipse_mid_time(self.period, self.sma_over_rs, self.eccentricity, self.inclination,
                                                 self.periastron, self.mid_time)
        self.ww = ww
        self.albedo = albedo
        self.emissivity = emissivity

        self.filters = {}

        self.observations = {}

    # filter-dependent values

    def filter(self, filter_name):

        if filter_name in self.filters:
            return self.filters[filter_name]

        else:
            try:
                _get_filter(filter_name)
                ldc1, ldc2, ldc3, ldc4 = exotethys(self.stellar_logg, self.stellar_temperature, self.stellar_metallicity, filter_name,
                                                   method=self.ldc_method, stellar_model=self.ldc_stellar_model)
                fp = fp_over_fs(self.rp_over_rs, self.sma_over_rs, self.albedo, self.emissivity, self.stellar_temperature,
                                filter_name)
                print('Fp/Fs estimated using A={0}, e={1} for filter {2}.'.format(self.albedo, self.emissivity,
                                                                                  filter_name))

                self.filters[filter_name] = Filter(self.rp_over_rs, ldc1, ldc2, ldc3, ldc4, fp)

                return self.filters[filter_name]

            except PyLCInputError:
                raise PyLCInputError('Filter not available, you need to add {0} to the existing filters first.'.format
                                     (filter_name))

    def add_filter(self, filter_name, rp_over_rs, ldc1, ldc2, ldc3, ldc4, fp):

        self.filters[filter_name] = Filter(rp_over_rs, ldc1, ldc2, ldc3, ldc4, fp)

    # planet calculations

    def planet_orbit(self, time, time_format):
        time = self.target.convert_to_bjd_tdb(time, time_format)
        return planet_orbit(self.period, self.sma_over_rs, self.eccentricity, self.inclination, self.periastron,
                            self.mid_time, time, ww=self.ww)

    def planet_star_projected_distance(self, time, time_format):
        time = self.target.convert_to_bjd_tdb(time, time_format)
        return planet_star_projected_distance(self.period, self.sma_over_rs, self.eccentricity, self.inclination,
                                              self.periastron, self.mid_time, time)

    def planet_phase(self, time, time_format):
        time = np.array([self.target.convert_to_bjd_tdb(ff, time_format) for ff in time])
        return planet_phase(self.period, self.mid_time, time)

    def transit(self, time, time_format, filter_name, precision=3):
        filter_data = self.filter(filter_name)
        time = self.target.convert_to_bjd_tdb(time, time_format)

        return transit(filter_data.limb_darkening_coefficients,
                       filter_data.rp_over_rs, self.period, self.sma_over_rs, self.eccentricity,
                       self.inclination, self.periastron, self.mid_time, time,
                       method=self.ldc_method, precision=precision)

    def transit_integrated(self, time, time_format, exp_time, time_stamp, filter_name, max_sub_exp_time=10, precision=3):

        if time_stamp == 'start':
            time = np.array(time) + 0.5 * exp_time / (60.0 * 60.0 * 24.0)
        elif time_stamp == 'mid':
            time = np.array(time)
        elif time_stamp == 'end':
            time = np.array(time) - 0.5 * exp_time / (60.0 * 60.0 * 24.0)
        else:
            raise PyLCInputError(
                'Not acceptable time stamp {0}. Please choose between "mid", "start", "end".'.format(time_stamp))

        filter_data = self.filter(filter_name)
        time = self.target.convert_to_bjd_tdb(time, time_format)

        return transit_integrated(filter_data.limb_darkening_coefficients,
                                  filter_data.rp_over_rs, self.period, self.sma_over_rs, self.eccentricity,
                                  self.inclination, self.periastron, self.mid_time, time, exp_time=exp_time,
                                  max_sub_exp_time=max_sub_exp_time,
                                  method=self.ldc_method, precision=precision)

    def transit_duration(self, filter_name):
        filter_data = self.filter(filter_name)
        return transit_duration(filter_data.rp_over_rs,
                                self.period, self.sma_over_rs, self.eccentricity, self.inclination, self.periastron)

    def transit_depth(self, filter_name, precision=6):
        filter_data = self.filter(filter_name)
        return transit_depth(filter_data.limb_darkening_coefficients,
                             filter_data.rp_over_rs, self.period, self.sma_over_rs, self.eccentricity,
                             self.inclination, self.periastron,
                             method=self.ldc_method, precision=precision)

    def eclipse(self, time, time_format, filter_name, precision=3):
        filter_data = self.filter(filter_name)
        time = self.target.convert_to_bjd_tdb(time, time_format)

        return eclipse(filter_data.fp_over_fs, filter_data.rp_over_rs,
                       self.period, self.sma_over_rs, self.eccentricity,
                       self.inclination, self.periastron, self.eclipse_mid_time, time, precision=precision)

    def eclipse_integrated(self, time, time_format, exp_time, time_stamp, filter_name, max_sub_exp_time=10, precision=3):

        if time_stamp == 'start':
            time = np.array(time) + 0.5 * exp_time / (60.0 * 60.0 * 24.0)
        elif time_stamp == 'mid':
            time = np.array(time)
        elif time_stamp == 'end':
            time = np.array(time) - 0.5 * exp_time / (60.0 * 60.0 * 24.0)
        else:
            raise PyLCInputError(
                'Not acceptable time stamp {0}. Please choose between "mid", "start", "end".'.format(time_stamp))

        filter_data = self.filter(filter_name)
        time = self.target.convert_to_bjd_tdb(time, time_format)

        return eclipse_integrated(filter_data.fp_over_fs, filter_data.rp_over_rs,
                                  self.period, self.sma_over_rs, self.eccentricity,
                                  self.inclination, self.periastron, self.eclipse_mid_time, time, exp_time=exp_time,
                                  max_sub_exp_time=max_sub_exp_time, precision=precision)

    def eclipse_duration(self, filter_name):
        filter_data = self.filter(filter_name)
        return eclipse_duration(filter_data.rp_over_rs, self.period, self.sma_over_rs, self.eccentricity,
                                self.inclination, self.periastron)

    def eclipse_depth(self, filter_name, precision=6):
        filter_data = self.filter(filter_name)
        return eclipse_depth(filter_data.fp_over_fs, filter_data.rp_over_rs, self.period,
                             self.sma_over_rs, self.eccentricity, self.inclination, self.periastron,
                             precision=precision)

    # data fitting

    def add_observation(self, time, time_format, exp_time, time_stamp, flux, flux_unc, flux_format, filter_name):

        filter_data = self.filter(filter_name)

        original_data = {
            'time': time,
            'time_stamp': time_stamp,
            'time_format': time_format,
            'flux': flux,
            'flux_format': flux_format,
            'flux_unc': flux_unc,
            'exp_time': exp_time,
            'filter': filter_name
        }

        if time_stamp == 'start':
            time = np.array(time) + 0.5 * exp_time / (60.0 * 60.0 * 24.0)
        elif time_stamp == 'mid':
            time = np.array(time)
        elif time_stamp == 'end':
            time = np.array(time) - 0.5 * exp_time / (60.0 * 60.0 * 24.0)
        else:
            raise PyLCInputError(
                'Not acceptable time stamp {0}. Please choose between "mid", "start", "end".'.format(time_stamp))

        date = self.target.convert_to_jd(time[0], time_format)

        time = np.array([self.target.convert_to_bjd_tdb(ff, time_format) for ff in time])

        if flux_format == 'mag':
            flux_unc = np.abs(np.array(flux_unc) * (-0.921034 * np.exp(0.921034 * flux[0] - 0.921034 * np.array(flux))))
            flux = 10 ** ((flux[0] - np.array(flux)) / 2.5)
        elif flux_format == 'flux':
            flux = np.array(flux)
            flux_unc = np.array(flux_unc)
        else:
            raise PyLCInputError('Not acceptable flux format {0}. Please choose between "flux" and '
                                 '"mag".'.format(flux_format))

        obs_id = 0
        while obs_id in self.observations:
            obs_id += 1

        check_transit = (np.mean(time) - self.mid_time) / self.period
        check_transit = abs(check_transit - int(check_transit))

        check_eclipse = (np.mean(time) - self.eclipse_mid_time) / self.period
        check_eclipse = abs(check_eclipse - int(check_eclipse))

        if check_transit < check_eclipse:
            observation_type = 'transit'
            epoch = int(round((np.mean(time) - self.mid_time) / self.period, 0))
        else:
            observation_type = 'eclipse'
            epoch = int(round((np.mean(time) - self.eclipse_mid_time) / self.period, 0))

        self.observations[obs_id] = {
            'target': self.name,
            'time': time,
            'dtime': time - time[0],
            'flux': flux,
            'flux_unc': flux_unc,
            'exp_time': exp_time,
            'filter': filter_name,
            'epoch': epoch,
            'date': date,
            'observation_type': observation_type,
            'original_data': original_data
        }

    def transit_fitting(self, output_folder,

                        max_sub_exp_time=10, precision=3,

                        detrending_order=2,

                        iterations=130000, walkers=200, burn_in=30000,

                        fit_ldc1=False, fit_ldc2=False, fit_ldc3=False, fit_ldc4=False,

                        fit_rp_over_rs=True, fit_individual_rp_over_rs=True,

                        fit_sma_over_rs=False, fit_inclination=False,

                        fit_mid_time=True, fit_period=False,
                        fit_individual_times=True,

                        fit_ldc_limits=[0.0, 1.0],
                        fit_rp_over_rs_limits=[0.5, 2.0],
                        fit_sma_over_rs_limits=[0.5, 2.0],
                        fit_inclination_limits=[70.0, 90.0],
                        fit_mid_time_limits=[-0.2, 0.2],
                        fit_period_limits=[0.8, 1.2],

                        counter='MCMC'):

        if not os.path.isdir(output_folder):
            os.mkdir(output_folder)

        parameters_map = [[] for observation in self.observations]

        names = []
        print_names = []
        limits1 = []
        limits2 = []
        initial = []

        # de-trending parameters

        for observation_num, observation in enumerate(self.observations):

            max_limit = (10 * (max(self.observations[observation]['flux']) - min(self.observations[observation]['flux'])) /
                         (max(self.observations[observation]['flux']) - min(self.observations[observation]['flux'])) / np.mean(self.observations[observation]['flux']))

            names.append('N_{0}'.format(observation_num + 1))
            print_names.append('N_{0}'.format(observation_num + 1))
            initial.append(np.mean(self.observations[observation]['flux']))
            limits1.append(0.5 * np.mean(self.observations[observation]['flux']))
            limits2.append(1.5 * np.mean(self.observations[observation]['flux']))

            parameters_map[observation_num].append(len(names) - 1)

            names.append('L_{0}'.format(observation_num + 1))
            print_names.append('L_{0}'.format(observation_num + 1))
            initial.append(0)
            if detrending_order >= 1:
                limits1.append(-max_limit)
                limits2.append(max_limit)
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            parameters_map[observation_num].append(len(names) - 1)

            names.append('Q_{0}'.format(observation_num + 1))
            print_names.append('Q_{0}'.format(observation_num + 1))
            initial.append(0)
            if detrending_order == 2:
                limits1.append(-5)
                limits2.append(5)
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            parameters_map[observation_num].append(len(names) - 1)

        # limb-darkening and rp_over_rs parameters

        unique_filters = []
        for observation in self.observations:
            if self.observations[observation]['filter'] not in unique_filters:
                unique_filters.append(self.observations[observation]['filter'])

        for phot_filter in unique_filters:

            filter_data = self.filter(phot_filter)
            ldc1, ldc2, ldc3, ldc4 = filter_data.limb_darkening_coefficients
            rp_over_rs = filter_data.rp_over_rs

            names.append('LDC1_{0}'.format(phot_filter))
            print_names.append('LDC1_{0}'.format(phot_filter))
            initial.append(ldc1)
            if fit_ldc1:
                limits1.append(fit_ldc_limits[0])
                limits2.append(fit_ldc_limits[1])
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            for observation_num, observation in enumerate(self.observations):
                if self.observations[observation]['filter'] == phot_filter:
                    parameters_map[observation_num].append(len(names) - 1)

            names.append('LDC2_{0}'.format(phot_filter))
            print_names.append('LDC2_{0}'.format(phot_filter))
            initial.append(ldc2)
            if fit_ldc2:
                limits1.append(fit_ldc_limits[0])
                limits2.append(fit_ldc_limits[1])
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            for observation_num, observation in enumerate(self.observations):
                if self.observations[observation]['filter'] == phot_filter:
                    parameters_map[observation_num].append(len(names) - 1)

            names.append('LDC3_{0}'.format(phot_filter))
            print_names.append('LDC3_{0}'.format(phot_filter))
            initial.append(ldc3)
            if fit_ldc3:
                limits1.append(fit_ldc_limits[0])
                limits2.append(fit_ldc_limits[1])
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            for observation_num, observation in enumerate(self.observations):
                if self.observations[observation]['filter'] == phot_filter:
                    parameters_map[observation_num].append(len(names) - 1)

            names.append('LDC4_{0}'.format(phot_filter))
            print_names.append('LDC4_{0}'.format(phot_filter))
            initial.append(ldc4)
            if fit_ldc4:
                limits1.append(fit_ldc_limits[0])
                limits2.append(fit_ldc_limits[1])
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            for observation_num, observation in enumerate(self.observations):
                if self.observations[observation]['filter'] == phot_filter:
                    parameters_map[observation_num].append(len(names) - 1)

            if fit_individual_rp_over_rs:

                names.append('rp_{0}'.format(phot_filter))
                print_names.append('(R_\mathrm{p}/R_*)_{' + phot_filter + '}')
                initial.append(rp_over_rs)
                if fit_rp_over_rs:
                    limits1.append(rp_over_rs * fit_rp_over_rs_limits[0])
                    limits2.append(rp_over_rs * fit_rp_over_rs_limits[1])
                else:
                    limits1.append(np.nan)
                    limits2.append(np.nan)

                for observation_num, observation in enumerate(self.observations):
                    if self.observations[observation]['filter'] == phot_filter:
                        parameters_map[observation_num].append(len(names) - 1)

        if not fit_individual_rp_over_rs:

            names.append('rp')
            print_names.append('(R_\mathrm{p}/R_*)')
            initial.append(self.rp_over_rs)
            if fit_rp_over_rs:
                limits1.append(self.rp_over_rs * fit_rp_over_rs_limits[0])
                limits2.append(self.rp_over_rs * fit_rp_over_rs_limits[1])
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            for observation_num, observation in enumerate(self.observations):
                parameters_map[observation_num].append(len(names) - 1)

        # orbital parameters

        names.append('P')
        print_names.append('P')
        initial.append(self.period)
        if fit_period:
            limits1.append(self.period * fit_period_limits[0])
            limits2.append(self.period * fit_period_limits[1])
        else:
            limits1.append(np.nan)
            limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        names.append('a')
        print_names.append('a/R_*')
        initial.append(self.sma_over_rs)
        if fit_sma_over_rs:
            limits1.append(self.sma_over_rs * fit_sma_over_rs_limits[0])
            limits2.append(self.sma_over_rs * fit_sma_over_rs_limits[1])
        else:
            limits1.append(np.nan)
            limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        names.append('e')
        print_names.append('e')
        initial.append(self.eccentricity)
        limits1.append(np.nan)
        limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        names.append('i')
        print_names.append('i')
        initial.append(self.inclination)
        if fit_inclination:
            limits1.append(fit_inclination_limits[0])
            limits2.append(fit_inclination_limits[1])
        else:
            limits1.append(np.nan)
            limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        names.append('w')
        print_names.append('\omega')
        initial.append(self.periastron)
        limits1.append(np.nan)
        limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        # time parameters

        test_epochs = []
        test_epochs_weights = []

        for observation in self.observations:
            test_epochs.append((self.observations[observation]['time'] - self.mid_time) / self.period)
            norm_errors = self.observations[observation]['flux_unc'] / self.observations[observation]['flux']
            test_epochs_weights.append(1 / (norm_errors * norm_errors))

        test_epochs = np.concatenate(test_epochs)
        test_epochs_weights = np.concatenate(test_epochs_weights)

        new_epoch = np.round(np.sum(test_epochs * test_epochs_weights) / np.sum(test_epochs_weights), 0)
        new_mid_time = self.mid_time + new_epoch * self.period

        for observation in self.observations:
            self.observations[observation]['epoch'] = int(round((np.mean(self.observations[observation]['time']) -new_mid_time) / self.period, 0))

        unique_epochs = []
        for observation in self.observations:
            if self.observations[observation]['epoch'] not in unique_epochs:
                unique_epochs.append(self.observations[observation]['epoch'])

        if not fit_individual_times:

            names.append('T_0')
            print_names.append('T_0')
            initial.append(new_mid_time)
            if fit_mid_time:
                limits1.append(new_mid_time + fit_mid_time_limits[0])
                limits2.append(new_mid_time + fit_mid_time_limits[1])
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            for observation_num, observation in enumerate(self.observations):
                parameters_map[observation_num].append(len(names) - 1)

        else:

            if fit_period:
                raise PyLCInputError('Period and individual mid times cannot be fitted simultaneously.')

            for epoch in unique_epochs:

                names.append('T_mid_{0}'.format(epoch))
                print_names.append('T_\mathrm{mid_' + str(epoch) + '}')
                initial.append(new_mid_time + epoch * self.period)
                if fit_mid_time:
                    limits1.append(new_mid_time + epoch * self.period + fit_mid_time_limits[0])
                    limits2.append(new_mid_time + epoch * self.period + fit_mid_time_limits[1])
                else:
                    limits1.append(np.nan)
                    limits2.append(np.nan)

                for observation_num, observation in enumerate(self.observations):
                    if self.observations[observation]['epoch'] == epoch:
                        parameters_map[observation_num].append(len(names) - 1)

        model_time = np.array([])
        model_flux = np.array([])
        model_flux_unc = np.array([])
        for observation in self.observations:
            model_time = np.append(model_time, self.observations[observation]['time'])
            model_flux = np.append(model_flux, self.observations[observation]['flux'])
            model_flux_unc = np.append(model_flux_unc, self.observations[observation]['flux_unc'])

        parameters_map = np.array(parameters_map)

        def detrend_model(model_time, *model_variables):

            model = []

            for observation_num, observation in enumerate(self.observations):

                detrend_zero,  detrend_one, detrend_two, ldc1, ldc2, ldc3, ldc4, r, p, a, e, i, w, mt = \
                    np.array(model_variables)[parameters_map[observation_num]]

                deltatime = self.observations[observation]['dtime']

                model.append(detrend_zero * (1 + detrend_one * deltatime + detrend_two * deltatime * deltatime))

            return np.concatenate(model)

        def full_model(model_time, *model_variables):

            model_variables = np.array(model_variables)

            model = []

            for observation_num, observation in enumerate(self.observations):
                detrend_zero, detrend_one, detrend_two, ldc1, ldc2, ldc3, ldc4, r, p, a, e, i, w, mt = \
                    model_variables[parameters_map[observation_num]]

                deltatime = self.observations[observation]['dtime']

                model1 = detrend_zero * (1 + detrend_one * deltatime + detrend_two * deltatime * deltatime)

                model.append(model1 * transit_integrated(
                    [ldc1, ldc2, ldc3, ldc4], r, p, a, e, i, w, mt,
                    time_array=self.observations[observation]['time'],
                    exp_time=self.observations[observation]['exp_time'],
                    max_sub_exp_time=max_sub_exp_time,
                    method=self.ldc_method,
                    precision=precision
                ))

            return np.concatenate(model)

        fitting = EmceeFitting(model_time, model_flux, model_flux_unc,
                               full_model, initial, limits1, limits2, walkers,
                               iterations, burn_in,
                               data_x_name='time', data_y_name='flux', data_x_print_name='Time',
                               data_y_print_name='Relative Flux',
                               parameters_names=names, parameters_print_names=print_names,
                               counter=counter)

        fitting.run_mcmc()
        results = fitting.results

        # TODO
        # save observations separately and calculate some statistics individually

        results['settings'] = {}
        results['settings']['max_sub_exp_time'] = max_sub_exp_time
        results['settings']['precision'] = precision
        results['settings']['detrending_order'] = detrending_order
        results['settings']['iterations'] = iterations
        results['settings']['walkers'] = walkers
        results['settings']['burn_in'] = burn_in
        results['settings']['fit_ldc1'] = fit_ldc1
        results['settings']['fit_ldc2'] = fit_ldc2
        results['settings']['fit_ldc3'] = fit_ldc3
        results['settings']['fit_ldc4'] = fit_ldc4
        results['settings']['fit_rp_over_rs'] = fit_rp_over_rs
        results['settings']['force_same_rp_over_rs'] = fit_individual_rp_over_rs
        results['settings']['fit_sma_over_rs'] = fit_sma_over_rs
        results['settings']['fit_inclination'] = fit_inclination
        results['settings']['fit_mid_time'] = fit_mid_time
        results['settings']['fit_period'] = fit_period
        results['settings']['fit_individual_times'] = fit_individual_times
        results['settings']['fit_ldc_limits'] = fit_ldc_limits
        results['settings']['fit_rp_over_rs_limits'] = fit_rp_over_rs_limits
        results['settings']['fit_sma_over_rs_limits'] = fit_sma_over_rs_limits
        results['settings']['fit_inclination_limits'] = fit_inclination_limits
        results['settings']['fit_mid_time_limits'] = fit_mid_time_limits
        results['settings']['fit_period_limits'] = fit_period_limits
        results['settings']['filter_map'] = self.filters

        results['detrended_input_series'] = {
            'time': results['input_series']['time'],
            'flux': results['input_series']['flux'] / detrend_model(model_time, *results['parameters_final']),
            'flux_unc': results['input_series']['flux_unc'] / detrend_model(model_time, *results['parameters_final'])}

        results['detrended_output_series'] = {
            'model': results['output_series']['model'] / detrend_model(model_time, *results['parameters_final']),
            'residuals': (results['output_series']['residuals']
                          / detrend_model(model_time, *results['parameters_final']))}

        results['detrended_statistics'] = {
            'res_mean': np.mean(results['detrended_output_series']['residuals']),
            'res_std': np.std(results['detrended_output_series']['residuals']),
            'res_rms': np.sqrt(np.mean(results['detrended_output_series']['residuals'] ** 2))
        }

        observations_id_series = np.array([])
        for observation in self.observations:
            observations_id_series = np.append(observations_id_series,
                                               np.ones(len(self.observations[observation]['time'])) * observation)

        for observation in self.observations:

            id_series = np.where(observations_id_series == observation)

            self.observations[observation]['model'] = results['output_series']['model'][id_series]
            self.observations[observation]['residuals'] = results['output_series']['residuals'][id_series]

            res_autocorr = np.correlate(self.observations[observation]['residuals'],
                                        self.observations[observation]['residuals'], mode='full')
            res_autocorr = res_autocorr[res_autocorr.size // 2:] / res_autocorr[res_autocorr.size // 2:][0]

            self.observations[observation]['res_autocorr'] = res_autocorr
            self.observations[observation]['res_max_autocorr'] = np.max(res_autocorr[1:])
            self.observations[observation]['res_mean'] = np.mean(self.observations[observation]['residuals'])
            self.observations[observation]['res_std'] = np.std(self.observations[observation]['residuals'])
            self.observations[observation]['res_rms'] = np.sqrt(np.mean(self.observations[observation]['residuals'] ** 2))
            self.observations[observation]['res_chi_sqr'] = np.sum(
                (self.observations[observation]['residuals'] ** 2) / (self.observations[observation]['flux_unc'] ** 2))
            self.observations[observation]['res_red_chi_sqr'] = (
                    self.observations[observation]['res_chi_sqr'] / (
                        len(self.observations[observation]['flux_unc']) - len(results['statistics']['corr_variables'])))

            self.observations[observation]['detrended_flux'] = results['detrended_input_series']['flux'][id_series]
            self.observations[observation]['detrended_flux_unc'] = results['detrended_input_series']['flux_unc'][id_series]
            self.observations[observation]['detrended_model'] = results['detrended_output_series']['model'][id_series]
            self.observations[observation]['detrended_residuals'] = results['detrended_output_series']['residuals'][id_series]

            res_autocorr = np.correlate(self.observations[observation]['detrended_residuals'],
                                        self.observations[observation]['detrended_residuals'], mode='full')
            res_autocorr = res_autocorr[res_autocorr.size // 2:] / res_autocorr[res_autocorr.size // 2:][0]

            self.observations[observation]['detrended_res_autocorr'] = res_autocorr
            self.observations[observation]['detrended_res_max_autocorr'] = np.max(res_autocorr[1:])
            self.observations[observation]['detrended_res_mean'] = np.mean(self.observations[observation]['detrended_residuals'])
            self.observations[observation]['detrended_res_std'] = np.std(self.observations[observation]['detrended_residuals'])
            self.observations[observation]['detrended_res_rms'] = np.sqrt(
                np.mean(self.observations[observation]['detrended_residuals'] ** 2))
            self.observations[observation]['detrended_res_chi_sqr'] = np.sum(
                (self.observations[observation]['detrended_residuals'] ** 2) / (self.observations[observation]['detrended_flux_unc'] ** 2))
            self.observations[observation]['detrended_res_red_chi_sqr'] = (
                    self.observations[observation]['detrended_res_chi_sqr'] / (
                    len(self.observations[observation]['detrended_flux_unc']) - len(results['statistics']['corr_variables'])))

            w = open(os.path.join(output_folder, 'diagnostics_dataset_{0}.txt'.format(observation + 1)), 'w')
            w.write('\n#Residuals:\n')
            w.write('#Mean: {0}\n'.format(self.observations[observation]['res_mean']))
            w.write('#STD: {0}\n'.format(self.observations[observation]['res_std']))
            w.write('#RMS: {0}\n'.format(self.observations[observation]['res_rms']))
            w.write('#Max auto-correlation: {0}\n'.format(self.observations[observation]['res_max_autocorr']))
            w.write('#Chi squared: {0}\n'.format(self.observations[observation]['res_chi_sqr']))
            w.write('#Reduced chi squared: {0}\n'.format(self.observations[observation]['res_red_chi_sqr']))

            w.write('\n\n#Detrended Residuals:\n')
            w.write('#Mean: {0}\n'.format(self.observations[observation]['detrended_res_mean']))
            w.write('#STD: {0}\n'.format(self.observations[observation]['detrended_res_std']))
            w.write('#RMS: {0}\n'.format(self.observations[observation]['detrended_res_rms']))
            w.write('#Max auto-correlation: {0}\n'.format(self.observations[observation]['detrended_res_max_autocorr']))
            w.write('#Chi squared: {0}\n'.format(self.observations[observation]['detrended_res_chi_sqr']))
            w.write('#Reduced chi squared: {0}\n'.format(self.observations[observation]['detrended_res_red_chi_sqr']))

            w.close()

        results['observations'] = self.observations

        save_dict(results, os.path.join(output_folder, 'results.pickle'))

        fitting.save_results(os.path.join(output_folder, 'results.txt'))

        fitting.plot_corner(os.path.join(output_folder, 'correlations.pdf'))

        fitting.plot_traces(os.path.join(output_folder, 'traces.pdf'))

        plot_transit_fitting_models(results, os.path.join(output_folder, 'lightcurves.pdf'))

    def eclipse_fitting(self, output_folder,

                        max_sub_exp_time=10, precision=3,

                        detrending_order=2,

                        iterations=130000, walkers=200, burn_in=30000,

                        fit_fp_over_fs=True, fit_individual_fp_over_fs=False,
                        fit_rp_over_rs=False, fit_individual_rp_over_rs=False,

                        fit_sma_over_rs=False, fit_inclination=False,

                        fit_mid_time=False, fit_period=False,
                        fit_individual_times=False,

                        fit_rp_over_rs_limits=[0.5, 2.0],
                        fit_fp_over_fs_limits=[0.001, 1000.0],
                        fit_sma_over_rs_limits=[0.5, 2.0],
                        fit_inclination_limits=[70.0, 90.0],
                        fit_mid_time_limits=[-0.2, 0.2],
                        fit_period_limits=[0.8, 1.2],

                        counter='MCMC'):

        if not os.path.isdir(output_folder):
            os.mkdir(output_folder)

        parameters_map = [[] for observation in self.observations]

        names = []
        print_names = []
        limits1 = []
        limits2 = []
        initial = []

        # de-trending parameters

        for observation_num, observation in enumerate(self.observations):

            max_limit = (10 * (max(self.observations[observation]['flux']) - min(self.observations[observation]['flux'])) /
                         (max(self.observations[observation]['flux']) - min(self.observations[observation]['flux'])) / np.mean(self.observations[observation]['flux']))

            names.append('N_{0}'.format(observation_num + 1))
            print_names.append('N_{0}'.format(observation_num + 1))
            initial.append(np.mean(self.observations[observation]['flux']))
            limits1.append(0.5 * np.mean(self.observations[observation]['flux']))
            limits2.append(1.5 * np.mean(self.observations[observation]['flux']))

            parameters_map[observation_num].append(len(names) - 1)

            names.append('L_{0}'.format(observation_num + 1))
            print_names.append('L_{0}'.format(observation_num + 1))
            initial.append(0)
            if detrending_order >= 1:
                limits1.append(-max_limit)
                limits2.append(max_limit)
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            parameters_map[observation_num].append(len(names) - 1)

            names.append('Q_{0}'.format(observation_num + 1))
            print_names.append('Q_{0}'.format(observation_num + 1))
            initial.append(0)
            if detrending_order == 2:
                limits1.append(-5)
                limits2.append(5)
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            parameters_map[observation_num].append(len(names) - 1)

        # limb-darkening and rp_over_rs parameters

        unique_filters = []
        for observation in self.observations:
            if self.observations[observation]['filter'] not in unique_filters:
                unique_filters.append(self.observations[observation]['filter'])

        fp_over_fs = 0.00001
        for phot_filter in unique_filters:

            filter_data = self.filter(phot_filter)
            fp_over_fs = filter_data.fp_over_fs

            if fit_individual_fp_over_fs:

                names.append('fp_{0}'.format(phot_filter))
                print_names.append('(F_\mathrm{p}/F_*)_{' + phot_filter + '}')
                initial.append(fp_over_fs)
                if fit_fp_over_fs:
                    limits1.append(fp_over_fs * fit_fp_over_fs_limits[0])
                    limits2.append(fp_over_fs * fit_fp_over_fs_limits[1])
                else:
                    limits1.append(np.nan)
                    limits2.append(np.nan)

                for observation_num, observation in enumerate(self.observations):
                    if self.observations[observation]['filter'] == phot_filter:
                        parameters_map[observation_num].append(len(names) - 1)

        if not fit_individual_fp_over_fs:

            names.append('fp')
            print_names.append('(F_\mathrm{p}/F_*)')
            initial.append(fp_over_fs)
            if fit_fp_over_fs:
                limits1.append(fp_over_fs * fit_fp_over_fs_limits[0])
                limits2.append(fp_over_fs * fit_fp_over_fs_limits[1])
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            for observation_num, observation in enumerate(self.observations):
                parameters_map[observation_num].append(len(names) - 1)

        rp_over_rs = self.rp_over_rs
        for phot_filter in unique_filters:

            filter_data = self.filter(phot_filter)
            rp_over_rs = filter_data.rp_over_rs

            if fit_individual_rp_over_rs:

                names.append('rp_{0}'.format(phot_filter))
                print_names.append('(R_\mathrm{p}/R_*)_{' + phot_filter + '}')
                initial.append(rp_over_rs)
                if fit_rp_over_rs:
                    limits1.append(rp_over_rs * fit_rp_over_rs_limits[0])
                    limits2.append(rp_over_rs * fit_rp_over_rs_limits[1])
                else:
                    limits1.append(np.nan)
                    limits2.append(np.nan)

                for observation_num, observation in enumerate(self.observations):
                    if self.observations[observation]['filter'] == phot_filter:
                        parameters_map[observation_num].append(len(names) - 1)

        if not fit_individual_rp_over_rs:

            names.append('rp')
            print_names.append('(R_\mathrm{p}/R_*)')
            initial.append(rp_over_rs)
            if fit_rp_over_rs:
                limits1.append(rp_over_rs * fit_rp_over_rs_limits[0])
                limits2.append(rp_over_rs * fit_rp_over_rs_limits[1])
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            for observation_num, observation in enumerate(self.observations):
                parameters_map[observation_num].append(len(names) - 1)

        # orbital parameters

        names.append('P')
        print_names.append('P')
        initial.append(self.period)
        if fit_period:
            limits1.append(self.period * fit_period_limits[0])
            limits2.append(self.period * fit_period_limits[1])
        else:
            limits1.append(np.nan)
            limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        names.append('a')
        print_names.append('a/R_*')
        initial.append(self.sma_over_rs)
        if fit_sma_over_rs:
            limits1.append(self.sma_over_rs * fit_sma_over_rs_limits[0])
            limits2.append(self.sma_over_rs * fit_sma_over_rs_limits[1])
        else:
            limits1.append(np.nan)
            limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        names.append('e')
        print_names.append('e')
        initial.append(self.eccentricity)
        limits1.append(np.nan)
        limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        names.append('i')
        print_names.append('i')
        initial.append(self.inclination)
        if fit_inclination:
            limits1.append(fit_inclination_limits[0])
            limits2.append(fit_inclination_limits[1])
        else:
            limits1.append(np.nan)
            limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        names.append('w')
        print_names.append('\omega')
        initial.append(self.periastron)
        limits1.append(np.nan)
        limits2.append(np.nan)

        for observation_num, observation in enumerate(self.observations):
            parameters_map[observation_num].append(len(names) - 1)

        # time parameters

        unique_epochs = []
        for observation in self.observations:
            if self.observations[observation]['epoch'] not in unique_epochs:
                unique_epochs.append(self.observations[observation]['epoch'])

        self.eclipse_mid_time += min(unique_epochs) * self.period

        if not fit_individual_times:

            names.append('T_0')
            print_names.append('T_0')
            initial.append(self.eclipse_mid_time)
            if fit_mid_time:
                limits1.append(self.eclipse_mid_time + fit_mid_time_limits[0])
                limits2.append(self.eclipse_mid_time + fit_mid_time_limits[1])
            else:
                limits1.append(np.nan)
                limits2.append(np.nan)

            for observation_num, observation in enumerate(self.observations):
                parameters_map[observation_num].append(len(names) - 1)

        else:

            if fit_period:
                raise PyLCInputError('Period and individual mid times cannot be fitted simultaneously.')

            for epoch in unique_epochs:

                epoch -= min(unique_epochs)

                names.append('T_{0}'.format(epoch))
                print_names.append('T_{' + str(epoch) + '}')
                initial.append(self.eclipse_mid_time + epoch * self.period)
                if fit_mid_time:
                    limits1.append(self.eclipse_mid_time + epoch * self.period + fit_mid_time_limits[0])
                    limits2.append(self.eclipse_mid_time + epoch * self.period + fit_mid_time_limits[1])
                else:
                    limits1.append(np.nan)
                    limits2.append(np.nan)

                for observation_num, observation in enumerate(self.observations):
                    if self.observations[observation]['epoch'] - min(unique_epochs) == epoch:
                        parameters_map[observation_num].append(len(names) - 1)

        model_time = np.array([])
        model_flux = np.array([])
        model_flux_unc = np.array([])
        for observation in self.observations:
            model_time = np.append(model_time, self.observations[observation]['time'])
            model_flux = np.append(model_flux, self.observations[observation]['flux'])
            model_flux_unc = np.append(model_flux_unc, self.observations[observation]['flux_unc'])

        parameters_map = np.array(parameters_map)

        def detrend_model(model_time, *model_variables):

            model = []

            for observation_num, observation in enumerate(self.observations):

                detrend_zero,  detrend_one, detrend_two, f, r, p, a, e, i, w, mt = \
                    np.array(model_variables)[parameters_map[observation_num]]

                deltatime = self.observations[observation]['dtime']

                model.append(detrend_zero * (1 + detrend_one * deltatime + detrend_two * deltatime * deltatime))

            return np.concatenate(model)

        def full_model(model_time, *model_variables):

            model = []

            for observation_num, observation in enumerate(self.observations):
                detrend_zero, detrend_one, detrend_two, f, r, p, a, e, i, w, mt = \
                    np.array(model_variables)[parameters_map[observation_num]]

                deltatime = self.observations[observation]['dtime']

                model1 = detrend_zero * (1 + detrend_one * deltatime + detrend_two * deltatime * deltatime)

                model.append(model1 * eclipse_integrated(f, r, p, a, e, i, w, mt,
                                                         time_array=self.observations[observation]['time'],
                                                         exp_time=self.observations[observation]['exp_time'],
                                                         max_sub_exp_time=max_sub_exp_time,
                                                         precision=precision
                                                         ))

            return np.concatenate(model)

        fitting = EmceeFitting(model_time, model_flux, model_flux_unc,
                               full_model, initial, limits1, limits2, walkers,
                               iterations, burn_in,
                               data_x_name='time', data_y_name='flux', data_x_print_name='Time',
                               data_y_print_name='Relative Flux',
                               parameters_names=names, parameters_print_names=print_names,
                               counter=counter)

        fitting.run_mcmc()
        results = fitting.results

        # TODO
        # save observations separately and calculate some statistics individually

        results['settings'] = {}
        results['settings']['max_sub_exp_time'] = max_sub_exp_time
        results['settings']['precision'] = precision
        results['settings']['detrending_order'] = detrending_order
        results['settings']['iterations'] = iterations
        results['settings']['walkers'] = walkers
        results['settings']['burn_in'] = burn_in
        results['settings']['fit_fp_over_fs'] = fit_fp_over_fs
        results['settings']['force_same_fp_over_fs'] = fit_individual_fp_over_fs
        results['settings']['fit_rp_over_rs'] = fit_rp_over_rs
        results['settings']['force_same_rp_over_rs'] = fit_individual_rp_over_rs
        results['settings']['fit_sma_over_rs'] = fit_sma_over_rs
        results['settings']['fit_inclination'] = fit_inclination
        results['settings']['fit_mid_time'] = fit_mid_time
        results['settings']['fit_period'] = fit_period
        results['settings']['fit_individual_times'] = fit_individual_times
        results['settings']['fit_rp_over_rs_limits'] = fit_rp_over_rs_limits
        results['settings']['fit_rp_over_rs_limits'] = fit_fp_over_fs_limits
        results['settings']['fit_sma_over_rs_limits'] = fit_sma_over_rs_limits
        results['settings']['fit_inclination_limits'] = fit_inclination_limits
        results['settings']['fit_mid_time_limits'] = fit_mid_time_limits
        results['settings']['fit_period_limits'] = fit_period_limits
        results['settings']['filter_map'] = self.filters

        results['detrended_input_series'] = {
            'time': results['input_series']['time'],
            'flux': results['input_series']['flux'] / detrend_model(model_time, *results['parameters_final']),
            'flux_unc': results['input_series']['flux_unc'] / detrend_model(model_time, *results['parameters_final'])}

        results['detrended_output_series'] = {
            'model': results['output_series']['model'] / detrend_model(model_time, *results['parameters_final']),
            'residuals': (results['output_series']['residuals']
                          / detrend_model(model_time, *results['parameters_final']))}

        results['detrended_statistics'] = {
            'res_mean': np.mean(results['detrended_output_series']['residuals']),
            'res_std': np.std(results['detrended_output_series']['residuals']),
            'res_rms': np.sqrt(np.mean(results['detrended_output_series']['residuals'] ** 2))
        }

        observations_id_series = np.array([])
        for observation in self.observations:
            observations_id_series = np.append(observations_id_series,
                                               np.ones(len(self.observations[observation]['time'])) * observation)

        for observation in self.observations:

            id_series = np.where(observations_id_series == observation)

            self.observations[observation]['model'] = results['output_series']['model'][id_series]
            self.observations[observation]['residuals'] = results['output_series']['residuals'][id_series]

            res_autocorr = np.correlate(self.observations[observation]['residuals'],
                                        self.observations[observation]['residuals'], mode='full')
            res_autocorr = res_autocorr[res_autocorr.size // 2:] / res_autocorr[res_autocorr.size // 2:][0]

            self.observations[observation]['res_autocorr'] = res_autocorr
            self.observations[observation]['res_max_autocorr'] = np.max(res_autocorr[1:])
            self.observations[observation]['res_mean'] = np.mean(self.observations[observation]['residuals'])
            self.observations[observation]['res_std'] = np.std(self.observations[observation]['residuals'])
            self.observations[observation]['res_rms'] = np.sqrt(np.mean(self.observations[observation]['residuals'] ** 2))
            self.observations[observation]['res_chi_sqr'] = np.sum(
                (self.observations[observation]['residuals'] ** 2) / (self.observations[observation]['flux_unc'] ** 2))
            self.observations[observation]['res_red_chi_sqr'] = (
                    self.observations[observation]['res_chi_sqr'] / (
                        len(self.observations[observation]['flux_unc']) - len(results['statistics']['corr_variables'])))

            self.observations[observation]['detrended_flux'] = results['detrended_input_series']['flux'][id_series]
            self.observations[observation]['detrended_flux_unc'] = results['detrended_input_series']['flux_unc'][id_series]
            self.observations[observation]['detrended_model'] = results['detrended_output_series']['model'][id_series]
            self.observations[observation]['detrended_residuals'] = results['detrended_output_series']['residuals'][id_series]

            res_autocorr = np.correlate(self.observations[observation]['detrended_residuals'],
                                        self.observations[observation]['detrended_residuals'], mode='full')
            res_autocorr = res_autocorr[res_autocorr.size // 2:] / res_autocorr[res_autocorr.size // 2:][0]

            self.observations[observation]['detrended_res_autocorr'] = res_autocorr
            self.observations[observation]['detrended_res_max_autocorr'] = np.max(res_autocorr[1:])
            self.observations[observation]['detrended_res_mean'] = np.mean(self.observations[observation]['detrended_residuals'])
            self.observations[observation]['detrended_res_std'] = np.std(self.observations[observation]['detrended_residuals'])
            self.observations[observation]['detrended_res_rms'] = np.sqrt(
                np.mean(self.observations[observation]['detrended_residuals'] ** 2))
            self.observations[observation]['detrended_res_chi_sqr'] = np.sum(
                (self.observations[observation]['detrended_residuals'] ** 2) / (self.observations[observation]['detrended_flux_unc'] ** 2))
            self.observations[observation]['detrended_res_red_chi_sqr'] = (
                    self.observations[observation]['detrended_res_chi_sqr'] / (
                    len(self.observations[observation]['detrended_flux_unc']) - len(results['statistics']['corr_variables'])))

            w = open(os.path.join(output_folder, 'diagnostics_dataset_{0}.txt'.format(observation + 1)), 'w')
            w.write('\n#Residuals:\n')
            w.write('#Mean: {0}\n'.format(self.observations[observation]['res_mean']))
            w.write('#STD: {0}\n'.format(self.observations[observation]['res_std']))
            w.write('#RMS: {0}\n'.format(self.observations[observation]['res_rms']))
            w.write('#Max auto-correlation: {0}\n'.format(self.observations[observation]['res_max_autocorr']))
            w.write('#Chi squared: {0}\n'.format(self.observations[observation]['res_chi_sqr']))
            w.write('#Reduced chi squared: {0}\n'.format(self.observations[observation]['res_red_chi_sqr']))

            w.write('\n\n#Detrended Residuals:\n')
            w.write('#Mean: {0}\n'.format(self.observations[observation]['detrended_res_mean']))
            w.write('#STD: {0}\n'.format(self.observations[observation]['detrended_res_std']))
            w.write('#RMS: {0}\n'.format(self.observations[observation]['detrended_res_rms']))
            w.write('#Max auto-correlation: {0}\n'.format(self.observations[observation]['detrended_res_max_autocorr']))
            w.write('#Chi squared: {0}\n'.format(self.observations[observation]['detrended_res_chi_sqr']))
            w.write('#Reduced chi squared: {0}\n'.format(self.observations[observation]['detrended_res_red_chi_sqr']))

            w.close()

        results['observations'] = self.observations

        save_dict(results, os.path.join(output_folder, 'results.pickle'))

        fitting.save_results(os.path.join(output_folder, 'results.txt'))

        fitting.plot_corner(os.path.join(output_folder, 'correlations.pdf'))

        fitting.plot_traces(os.path.join(output_folder, 'traces.pdf'))

        plot_transit_fitting_models(results, os.path.join(output_folder, 'lightcurves.pdf'))




__all__ = ['Degrees', 'Hours', 'Rad', 'arcsin', 'arccos', 'arctan', 'pi']


import numpy as np
from functools import lru_cache

from pylightcurve.errors import *


def _break_seconds(seconds):

    seconds = float(seconds)
    seconds_int = int(seconds)
    seconds_dec = int(str(seconds).split('.')[1])
    a = int(seconds_int / 3600.0)
    seconds_int = int(round(seconds_int - a * 3600.0))
    b = int(seconds_int / 60.0)
    seconds_int = int(round(seconds_int - b * 60.0))
    c = round(float('{0}.{1}'.format(seconds_int, seconds_dec)), 6)
    return a, b, c


def _collapse_to_seconds(degrees_hours, minutes, seconds):

        try:
            degrees_hours = float(degrees_hours)
        except:

            if isinstance(degrees_hours, str):

                if minutes != 0 or seconds != 0:
                    raise PyLCInputError('Not valid angle format')
                else:
                    try:
                        degrees_hours, minutes, seconds = degrees_hours.replace(':', ' ').split()
                        if degrees_hours[0] == '-':
                            minutes = '-{0}'.format(minutes)
                            seconds = '-{0}'.format(seconds)
                    except:
                        raise PyLCInputError('Not valid angle format')

        try:

            degrees_hours = float(degrees_hours)
            minutes = float(minutes)
            seconds = float(seconds)

        except:
            raise PyLCInputError('Not valid angle format')

        if degrees_hours <= 0 and minutes <= 0 and seconds <= 0:
            pass
        elif degrees_hours >= 0 and minutes >= 0 and seconds >= 0:
            pass
        else:
            raise PyLCInputError('Not valid angle format. Hours, degrees, minutes and seconds should be either ALL '
                                 'positive or ALL negative.')

        return round(degrees_hours * 3600.0 + minutes * 60.0 + seconds, 6)


class _DMS:

    def __init__(self, arcseconds):
        self.d, self.m, self.s = _break_seconds(arcseconds)
        self.list = [self.d, self.m, self.s]

    def __str__(self):
        return '{0}:{1}:{2}{3}'.format(str(self.d).zfill(2), str(self.m).zfill(2), '0'*(self.s < 10), str(self.s))

    def __repr__(self):
        return self.__str__()


class _HMS:

    def __init__(self, seconds):
        self.h, self.m, self.s = _break_seconds(seconds)

        self.list = [self.h, self.m, self.s]

    def __str__(self):
        return '{0}:{1}:{2}{3}'.format(str(self.h).zfill(2), str(self.m).zfill(2), '0'*(self.s < 10), str(self.s))

    def __repr__(self):
        return self.__str__()


class _Angle:

    def __init__(self, arcseconds):

        arcseconds = round(arcseconds, 6)

        if arcseconds < 0:
            arcseconds += 1296000.0 * (int(abs(arcseconds) / 1296000.0) + 1.0)
        arcseconds -= 1296000.0 * int(arcseconds / 1296000.0)

        self.arcseconds = arcseconds

        self._definition = '{0} arcsec'.format(arcseconds)

    @lru_cache()
    def deg(self):
        return self.arcseconds / 3600.0

    @lru_cache()
    def deg_coord(self):
        if self.deg() <= 90:
            return self.deg()

        elif self.deg() <= 180:
            return 180 - self.deg()

        elif self.deg() <= 270:
            return - (self.deg() - 180.0)

        else:
            return - (360.0 - self.deg())

    @lru_cache()
    def hours(self):
        return self.arcseconds / 15.0 / 3600.0

    @lru_cache()
    def rad(self):
        return (self.arcseconds / 3600.0) * np.pi / 180.0

    @lru_cache()
    def dms(self):
        return _DMS(self.arcseconds)

    @lru_cache()
    def dms_coord(self):

        if self.deg() <= 90:
            sign = '+'
            dec_print = self

        elif self.deg() <= 180:
            sign = '+'
            dec_print = pi - self

        elif self.deg() <= 270:
            sign = '-'
            dec_print = self - pi

        else:
            sign = '-'
            dec_print = 2 * pi - self

        return '{0}{1}'.format(sign, dec_print.dms())

    @lru_cache()
    def hms(self):
        return _HMS(self.arcseconds / 15.0)

    @lru_cache()
    def sin(self):
        return np.sin(self.rad())

    @lru_cache()
    def cos(self):
        return np.cos(self.rad())

    @lru_cache()
    def tan(self):
        return np.tan(self.rad())

    @lru_cache()
    def _get_class(self):
        return 'plc.{0}'.format(str(self.__class__).split('.')[-1][:-2])

    def __add__(self, other):
        _request_angle(other)
        return Degrees(0, 0, self.arcseconds + other.arcseconds)

    def __sub__(self, other):
        _request_angle(other)
        return Degrees(0, 0, self.arcseconds - other.arcseconds)

    def __mul__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return Degrees(0, 0, self.arcseconds * other)
        else:
            raise PyLCError('Operation not supported between {0} and {1}.'.format(self._get_class(), type(other)))

    def __truediv__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return Degrees(0, 0, self.arcseconds / other)
        else:
            raise PyLCError('Operation not supported between {0} and {1}.'.format(self._get_class(), type(other)))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return '{0}({1} DMS, defined as {2})'.format(self._get_class(), self.dms(), self._definition)


class Degrees(_Angle):

    def __init__(self, degrees, arcminutes=0.0, arcseconds=0.0):

        total_arcseconds = _collapse_to_seconds(degrees, arcminutes, arcseconds)

        _Angle.__init__(self, total_arcseconds)

        if arcminutes == 0 and arcseconds == 0:
            self._definition = '{0} degrees'.format(degrees)
        else:
            self._definition = '{0} degrees, {1} arcminutes, {2} arcseconds'.format(degrees, arcminutes, arcseconds)


class Hours(_Angle):

    def __init__(self, hours, minutes=0.0, seconds=0.0):

        total_arcseconds = _collapse_to_seconds(hours, minutes, seconds) * 15.0

        _Angle.__init__(self, total_arcseconds)

        if minutes == 0 and seconds == 0:
            self._definition = '{0} hours'.format(hours)
        else:
            self._definition = '{0} hours, {1} minutes, {2} seconds'.format(hours, minutes, seconds)


class Rad(_Angle):

    def __init__(self, rad):

        try:
            arcseconds = float(rad) * 648000.0 / np.pi
        except:
            raise PyLCInputError('Not valid input for rad.')

        _Angle.__init__(self, arcseconds)

        self._definition = '{0} rad'.format(rad)


def _is_angle(obsject):
    if isinstance(obsject, Degrees) or isinstance(obsject, Hours) or isinstance(obsject, Rad):
        return True
    else:
        return False


def _request_angle(item):
    if _is_angle(item):
        pass
    else:
        raise PyLCInputError('An angle object is required (plc.Degrees, plc.Hours or plc.Rad)')


def arccos(number):
    return Rad(np.arccos(number))


def arcsin(number):
    return Rad(np.arcsin(number))


def arctan(number):
    return Rad(np.arctan(number))


pi = Degrees(180.0)





class PyLCError(BaseException):
    pass


class PyLCLibraryError(PyLCError):
    pass


class PyLCFileError(PyLCError):
    pass


class PyLCProcessError(PyLCError):
    pass


class PyLCInputError(PyLCError):
    pass




import numpy as np

from pylightcurve.errors import *
from pylightcurve.analysis.distributions import two_d_distribution

import matplotlib.cm as cm
import matplotlib.gridspec as gridspec
from matplotlib.figure import Figure
from matplotlib.backend_bases import FigureCanvasBase


# mcmc plots


def plot_mcmc_fitting(fitting_object, output_file):

    if not fitting_object.mcmc_run_complete:
        raise PyLCProcessError('MCMC not completed')

    results = fitting_object.results

    f1 = 3
    f2 = 1

    fbottom = 0.12
    ftop = 0.05
    fleft = 0.1
    fright = 0.05

    funit = 1.0
    frow = f1 + f2
    fcol = round(frow * 1.5)
    frow_height = (1.0 - fbottom - ftop) / frow

    fwidth = funit * fcol / (1 - fright - fleft)
    fheight = funit * frow / (1 - fbottom - ftop)

    lebels_right = fleft / 4

    fsmain = 2.0 * fheight
    fsbig = 1.5 * fsmain

    fig = Figure(figsize=(fwidth, fheight))
    canvas = FigureCanvasBase(fig)
    try:
        gs = gridspec.GridSpec(frow, fcol, fig, fleft, fbottom, 1.0 - fright, 1.0 - ftop, 0.0, 0.0)
    except TypeError:
        gs = gridspec.GridSpec(frow, fcol, fleft, fbottom, 1.0 - fright, 1.0 - ftop, 0.0, 0.0)

    # raw
    ax1 = fig.add_subplot(gs[0:f1, 0:])

    ax1.errorbar(results['input_series'][results['input_series_x']],
                 results['input_series'][results['input_series_y']],
                 results['input_series'][results['input_series_y_unc']],
                 fmt='o', color='k', ms=2)
    ax1.plot(results['input_series'][results['input_series_x']],
             results['output_series']['model'], 'r-')

    fig.text(lebels_right, fbottom + (f2 + f1 / 2) * frow_height, results['input_series_y_print'],
             fontsize=fsbig, va='center', ha='center', rotation='vertical')

    ax1.tick_params(labelbottom=False, labelsize=fsmain)

    # residuals
    ax2 = fig.add_subplot(gs[f1:f1 + f2, 0:])
    ax2.errorbar(results['input_series'][results['input_series_x']],
                 results['output_series']['residuals'],
                 results['input_series'][results['input_series_y_unc']],
                 fmt='o', color='k', ms=2)
    ax2.plot(results['input_series'][results['input_series_x']],
             results['output_series']['residuals'] * 0, 'r-')

    ax2.set_ylim(- 8 * results['statistics']['res_std'], 8 * results['statistics']['res_std'])

    ax2.set_xlabel(results['input_series_x_print'], fontsize=fsbig)
    fig.text(lebels_right, fbottom + (f2 / 2) * frow_height, 'residuals', fontsize=fsbig, va='center', ha='center',
             rotation='vertical')

    ax2.tick_params(labelsize=fsmain)

    fig.savefig(output_file, transparent=False)
    del fig


def plot_mcmc_corner(fitting_object, export_file):

    if not fitting_object.mcmc_run_complete:
        raise PyLCProcessError('MCMC not completed')

    names = []
    results = []
    print_results = []
    errors1 = []
    print_errors1 = []
    errors2 = []
    print_errors2 = []
    errors = []
    traces = []
    traces_bins = []
    traces_counts = []

    for i in fitting_object.names:
        if fitting_object.results['parameters'][i]['initial']:
            names.append(fitting_object.results['parameters'][i]['print_name'])
            results.append(fitting_object.results['parameters'][i]['value'])
            print_results.append(fitting_object.results['parameters'][i]['print_value'])
            errors1.append(fitting_object.results['parameters'][i]['m_error'])
            print_errors1.append(fitting_object.results['parameters'][i]['print_m_error'])
            errors2.append(fitting_object.results['parameters'][i]['p_error'])
            print_errors2.append(fitting_object.results['parameters'][i]['print_p_error'])
            errors.append(0.5 * (fitting_object.results['parameters'][i]['m_error'] +
                                 fitting_object.results['parameters'][i]['p_error']))
            traces.append(fitting_object.results['parameters'][i]['trace'])
            traces_bins.append(fitting_object.results['parameters'][i]['trace_bins'])
            traces_counts.append(fitting_object.results['parameters'][i]['trace_counts'])

    correlation = fitting_object.results['statistics']['corr_matrix']


    all_var = len(traces)

    fig = Figure(figsize=(2.5 * all_var + 0.5, 2.5 * all_var + 0.5))
    canvas = FigureCanvasBase(fig)
    cmap = cm.get_cmap('brg')
    try:
        gs = gridspec.GridSpec(all_var, all_var, fig, 0.5 / (2.5 * all_var + 0.5), 0.85 / (2.5 * all_var + 0.5),
                               1 - 0.5 / (2.5 * all_var + 0.5), 1 - 0.15 / (2.5 * all_var + 0.5), 0.0, 0.0)
    except TypeError:
        gs = gridspec.GridSpec(all_var, all_var, 0.5 / (2.5 * all_var + 0.5), 0.85 / (2.5 * all_var + 0.5),
                               1 - 0.5 / (2.5 * all_var + 0.5), 1 - 0.15 / (2.5 * all_var + 0.5), 0.0, 0.0)

    fontsize = 20

    for var in range(len(names)):

        ax = fig.add_subplot(gs[var, var], facecolor='w')

        ax.step(traces_bins[var], traces_counts[var], color='k', where='mid')

        ax.axvline(results[var], c='k')
        ax.axvline(results[var] - errors1[var], c='k', ls='--', lw=0.5)
        ax.axvline(results[var] + errors2[var], c='k', ls='--', lw=0.5)

        ax.set_xticks([])
        ax.set_yticks([])
        ax.tick_params(left=False, right=False, top=False, bottom=False, labelbottom=False, labelleft=False)

        ax.set_xlabel('{0}{3}{0}\n{0}{4}_{1}-{5}{2}^{1}+{6}{2}{0}'.format(
            '$', '{', '}', names[var], print_results[var], print_errors1[var], print_errors2[var]), fontsize=fontsize)

        ax.set_xlim(results[var] - 6 * errors[var], results[var] + 6 * errors[var])
        ax.set_ylim(0, ax.get_ylim()[1])

        for j in range(var + 1, all_var):

            ax2 = fig.add_subplot(gs[var, j], facecolor='w')

            binsx, binsy, final = two_d_distribution(traces[j], traces[var])
            ax2.imshow(np.where(final > 0, np.log(np.where(final > 0, final, 1)), 0),
                       extent=(np.min(binsx), np.max(binsx), np.min(binsy), np.max(binsy)),
                       cmap=cm.Greys, origin='lower', aspect='auto')

            ax2.set_xticks([])
            ax2.set_yticks([])
            ax2.tick_params(bottom=False, left=False, right=False, top=False, labelbottom=False,
                            labelleft=False, labelright=False, labeltop=False)

            ax2.set_xlim(results[j] - 6 * errors[j], results[j] + 6 * errors[j])
            ax2.set_ylim(results[var] - 6 * errors[var], results[var] + 6 * errors[var])
            text_x = ax2.get_xlim()[1] - 0.05 * (ax2.get_xlim()[1] - ax2.get_xlim()[0])
            text_y = ax2.get_ylim()[1] - 0.05 * (ax2.get_ylim()[1] - ax2.get_ylim()[0])
            ax2.text(text_x, text_y, '{0}{1}{2}'.format(r'$', str(round(correlation[var][j], 2)), '$'),
                     color=cmap(abs(correlation[var][j]) / 2.),
                     fontsize=fontsize, ha='right', va='top')

    fig.savefig(export_file, transparent=False)
    del fig


def plot_mcmc_traces(fitting_object, export_file):

    if not fitting_object.mcmc_run_complete:
        raise PyLCProcessError('MCMC not completed')

    fig = Figure(figsize=(7, 2.5 * len(fitting_object.fitted_parameters)), tight_layout=False)
    canvas = FigureCanvasBase(fig)
    cmap = cm.get_cmap('brg')

    for var_num, var in enumerate(fitting_object.fitted_parameters):

        ax = fig.add_subplot(len(fitting_object.fitted_parameters), 1, var_num + 1)
        ax.plot(fitting_object.results['parameters'][var]['trace'], 'k-', lw=0.1)
        ax.axhline(fitting_object.results['parameters'][var]['value'], c='r')
        ax.axhline(fitting_object.results['parameters'][var]['value'] - fitting_object.results['parameters'][var]['m_error'],
                   ls='--', c='r', lw=0.5)
        ax.axhline(fitting_object.results['parameters'][var]['value'] + fitting_object.results['parameters'][var]['m_error'],
                   ls='--', c='r', lw=0.5)

        ax.set_yticks([])
        ax.tick_params(left=False, right=False, labelleft=False)

        ax.set_ylabel('{0} = {1}\n{2}\n{3}'.format(r'${0}$'.format(fitting_object.results['parameters'][var]['print_name']),
                                                  r'${0}$'.format(fitting_object.results['parameters'][var]['print_value']),
                                                  r'$-{0}$'.format(fitting_object.results['parameters'][var]['print_m_error']),
                                                  r'$+{0}$'.format(fitting_object.results['parameters'][var]['print_p_error'])),
                      fontsize=15)

        if var_num != len(fitting_object.fitted_parameters) - 1:
            ax.tick_params(labelbottom='off')
        else:
            ax.set_xlabel(r'$\mathrm{iteration}$', fontsize=13)

        ax.set_ylim(fitting_object.results['parameters'][var]['value'] - 6 * fitting_object.results['parameters'][var]['m_error'],
                    fitting_object.results['parameters'][var]['value'] + 6 * fitting_object.results['parameters'][var]['m_error'])

    fig.subplots_adjust(hspace=0, wspace=0)
    fig.savefig(export_file, transparent=False)
    del fig


def plot_transit_fitting_models(results, output_file):

    for observation in results['observations']:

        data = results['observations'][observation]

        f1 = 3
        f2 = 3
        f3 = 1

        fbottom = 0.08
        ftop = 0.05
        fleft = 0.1
        fright = 0.05

        funit = 1.0
        frow = f1 + f2 + f3
        fcol = round(frow * 1.5)
        frow_height = (1.0 - fbottom - ftop) / frow

        lebels_right = fleft / 4

        fsmain = 10
        fsbig = 15
        fig = Figure(figsize=(funit * fcol / (1 - fright - fleft), funit * frow / (1 - fbottom - ftop)))
        canvas = FigureCanvasBase(fig)
        try:
            gs = gridspec.GridSpec(frow, fcol, fig, fleft, fbottom, 1.0 - fright, 1.0 - ftop, 0.0, 0.0)
        except TypeError:
            gs = gridspec.GridSpec(frow, fcol, fleft, fbottom, 1.0 - fright, 1.0 - ftop, 0.0, 0.0)

        fig.text(lebels_right, 1 - ftop / 2, 'relative flux', fontsize=fsbig, va='center', ha='left')

        # raw
        ax1 = fig.add_subplot(gs[0:f1, 0:])

        ax1.plot(data['time'], data['flux'], 'ko', ms=2)
        ax1.plot(data['time'], data['model'], 'r-', lw=1)

        fig.text(lebels_right, fbottom + (f3 + f2 + f1 / 2) * frow_height, 'raw', fontsize=fsbig, va='center',
                 ha='center', rotation='vertical')

        data_ymin = (min(data['flux']) - 3 * np.std(data['residuals']))

        data_ymax = (max(data['flux']) + 2 * np.std(data['residuals']))

        ax1.set_yticks(ax1.get_yticks()[np.where(ax1.get_yticks() > data_ymin)])

        ymin, ymax = data_ymax - 1.05 * (data_ymax - data_ymin), data_ymax

        ax1.set_ylim(ymin, ymax)

        ax1.tick_params(labelbottom=False, labelsize=fsmain)

        # de-trended
        ax2 = fig.add_subplot(gs[f1:f1+f2, 0:])

        ax2.plot(data['time'], data['detrended_flux'], 'ko', ms=2)
        ax2.plot(data['time'], data['detrended_model'], 'r-', lw=1)

        fig.text(lebels_right, fbottom + (f3 + f2 / 2) * frow_height, 'de-trended', fontsize=fsbig, va='center',
                 ha='center', rotation='vertical')

        data_ymin = (min(data['detrended_flux']) - 3 * np.std(data['detrended_residuals']))

        data_ymax = (max(data['detrended_flux']) + 2 * np.std(data['detrended_residuals']))

        ax2.set_yticks(ax2.get_yticks()[np.where(ax2.get_yticks() > data_ymin)])

        ymin, ymax = data_ymax - 1.05 * (data_ymax - data_ymin), data_ymax

        ax2.set_ylim(ymin, ymax)

        ax2.tick_params(labelbottom=False, labelsize=fsmain)

        # residuals
        ax3 = fig.add_subplot(gs[f1 + f2:f1 + f2 + f3, 0:])
        ax3.plot(data['time'], data['detrended_residuals'], 'ko', ms=2)
        ax3.plot(data['time'], np.zeros_like(data['time']), 'r-', lw=1)

        ax3.set_ylim(- 8 * np.std(data['detrended_residuals']),
                     8 * np.std(data['detrended_residuals']))

        ax3.set_xlabel('t [BJD_$\mathrm{TDB}$]', fontsize=fsbig)
        fig.text(lebels_right, fbottom + (f3 / 2) * frow_height, 'residuals', fontsize=fsbig, va='center', ha='center',
                 rotation='vertical')

        ax3.tick_params(labelsize=fsmain)

        ax1.set_title('{0} - Dataset {1} - {2}'.format(data['target'], observation + 1, data['date']))

        fig.savefig('.'.join(output_file.split('.')[:-1]) + '_dataset_{0}.'.format(observation + 1) + output_file.split('.')[-1])





__all__ = ['find_single_star']

import numpy as np
import warnings

from pylightcurve.analysis.gaussian import fit_two_d_gaussian
from pylightcurve.analysis.distributions import one_d_distribution


def find_single_star(data_array, predicted_x, predicted_y, mean=None, std=None, burn_limit=65000, star_std=2,
                     std_limit=5.0):
    star = None

    if 0 < predicted_x < len(data_array[0]) and 0 < predicted_x < len(data_array):

        if mean is None or std is None:
            fit_mean, fit_std = one_d_distribution(data_array, gaussian_fit=True, mad_filter=5)[2:4]

            if not mean:
                mean = fit_mean

            if not std:
                std = fit_std

        centroids = find_centroids(data_array, predicted_x - 5 * star_std, predicted_x + 5 * star_std,
                                   predicted_y - 5 * star_std, predicted_y + 5 * star_std, mean, std, burn_limit, star_std,
                                   std_limit)

        centroids = sorted(centroids, key=lambda x: np.sqrt((x[0] - predicted_x) ** 2 + (x[1] - predicted_y) ** 2))

        for centroid in centroids:
            star = _star_from_centroid(data_array, centroid[0], centroid[1], mean, std, burn_limit, star_std, std_limit)
            if star:
                star = [star[0][2], star[0][3], star[0][0], star[0][1], star[0][4], star[0][5], centroid[0], centroid[1]]
                break

    return star


def _star_from_centroid(data_array, centroid_x, centroid_y, mean, std, burn_limit, star_std, std_limit):

    star = None
    try:
        search_window = int(round(10 * star_std))
        y_min = int(max(int(centroid_y) - search_window, 0))
        y_max = int(min(int(centroid_y) + search_window, len(data_array) - 1))
        x_min = int(max(int(centroid_x) - search_window, 0))
        x_max = int(min(int(centroid_x) + search_window, len(data_array[0]) - 1))

        datax, datay = np.meshgrid(np.arange(x_min, x_max + 1) + 0.5,
                                   np.arange(y_min, y_max + 1) + 0.5)

        dataz = data_array[y_min: y_max + 1, x_min: x_max + 1]
        popt, pcov = fit_two_d_gaussian(datax, datay, dataz, positive=True, point_xy=(centroid_x, centroid_y),
                                        sigma=star_std, maxfev=1000)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if popt[0] > std_limit * std and popt[0] + popt[1] < burn_limit:
                if np.sqrt(pcov[0][0]) != np.inf:
                    if popt[0] > std_limit * np.sqrt(pcov[0][0]):
                        star = (popt, pcov)
                else:
                    star = (popt, pcov)
    except:
        pass

    return star


def find_centroids(data_array, x_low, x_upper, y_low, y_upper, mean, std, burn_limit, star_std, std_limit):

    x_upper = int(min(x_upper, len(data_array[0])))
    y_upper = int(min(y_upper, len(data_array)))
    x_low = int(max(0, x_low))
    y_low = int(max(0, y_low))

    data_array = np.full_like(data_array[y_low:y_upper + 1, x_low:x_upper + 1],
                              data_array[y_low:y_upper + 1, x_low:x_upper + 1])

    test = []

    for i in range(-star_std, star_std + 1):
        for j in range(-star_std, star_std + 1):
            rolled = np.roll(np.roll(data_array, i, 0), j, 1)
            test.append(rolled)

    median_test = np.median(test, 0)
    max_test = np.max(test, 0)
    del test
    stars = np.where((data_array < burn_limit) & (data_array > mean + std_limit * std) & (max_test == data_array)
                     & (median_test > mean + 2 * std))
    del data_array

    stars = [stars[1] + x_low, stars[0] + y_low]
    stars = np.swapaxes(stars, 0, 1)

    return stars




__all__ = ['gaussian', 'fit_gaussian', 'two_d_gaussian', 'fit_two_d_gaussian']


import numpy as np

from pylightcurve.analysis.curve_fit import curve_fit
from pylightcurve.analysis.numerical_integration import sample_function


# gaussian 1D


def gaussian(x_array, model_norm, model_floor, model_mean, model_std):

    return model_floor + (model_norm *
                          np.exp(- 0.5 * (model_mean - x_array) * (model_mean - x_array) / (model_std * model_std)))


def fit_gaussian(datax, datay, positive=False, sampled=False, sampled_precision=3, maxfev=10000):

    # TODO option to restrict the space searched

    datax = np.array(datax, dtype=np.float)
    datay = np.array(datay, dtype=np.float)

    # TODO option to point towards the solution

    initial_floor = np.median(datay)
    initial_norm = np.max(datay) - initial_floor
    initial_mean = datax[np.argmax(datay)]
    initial_sigma = np.abs(datax[np.argmin(np.abs(datay - np.max(datay) / 2))] - initial_mean)

    initial_values = [initial_norm, initial_floor, initial_mean, initial_sigma]

    if positive:
        def gaussian_to_fit(x_array, model_norm, model_floor, model_mean, model_std):
            return gaussian(x_array, np.abs(model_norm), model_floor, model_mean, model_std)
    else:
        gaussian_to_fit = gaussian

    if sampled:
        splits = 0.5 * (datax[1:] + datax[:-1])
        datax1 = np.array([datax[0] - (splits[0] - datax[0])] + list(splits))
        datax2 = np.array(list(splits) + [datax[-1] + (datax[-1] - splits[-1])])

        datax = (datax1, datax2)
        gaussian_to_fit = sample_function(gaussian_to_fit, precision=sampled_precision)

    popt, pcov = curve_fit(gaussian_to_fit, datax, datay, p0=initial_values, maxfev=maxfev)

    popt[3] = np.abs(popt[3])
    if positive:
        popt[0] = np.abs(popt[0])
        popt[1] = np.abs(popt[1])

    return popt, pcov


# gaussian 2D

def two_d_gaussian(x_array, y_array, model_norm, model_floor, model_x_mean, model_y_mean, model_x_sigma, model_y_sigma,
                   model_theta):

    a = (np.cos(model_theta) ** 2) / (2 * model_x_sigma ** 2) + (np.sin(model_theta) ** 2) / (2 * model_y_sigma ** 2)
    b = -(np.sin(2 * model_theta)) / (4 * model_x_sigma ** 2) + (np.sin(2 * model_theta)) / (4 * model_y_sigma ** 2)
    c = (np.sin(model_theta) ** 2) / (2 * model_x_sigma ** 2) + (np.cos(model_theta) ** 2) / (2 * model_y_sigma ** 2)

    return (model_floor + model_norm * np.exp(- (a * ((x_array - model_x_mean) ** 2)
                                              + 2.0 * b * (x_array - model_x_mean) * (y_array - model_y_mean)
                                              + c * ((y_array - model_y_mean) ** 2))))


def fit_two_d_gaussian(datax, datay, dataz, positive=False, symmetric=False, point_xy=None, sigma=None,
                       maxfev=10000):

    datax = np.array(datax, dtype=np.float)
    datay = np.array(datay, dtype=np.float)
    dataz = np.array(dataz, dtype=np.float)

    initial_floor = np.median(dataz)
    initial_norm = np.max(dataz) - initial_floor
    if sigma is not None:
        initial_x_sigma = sigma
        initial_y_sigma = sigma
    else:
        initial_x_sigma = 1.0
        initial_y_sigma = 1.0

    if not point_xy:
        test_data_x = np.sum(dataz, 0)
        test_floor = np.median(test_data_x)
        test_norm = np.max(test_data_x) - test_floor

        initial_x_mean_arg = int(len(datax[0]) / 2)
        model = gaussian(datax[0], test_norm, test_floor, datax[0][initial_x_mean_arg], initial_x_sigma)
        dx = np.argmax(np.convolve(test_data_x, model)) - np.argmax(np.convolve(model, model))
        initial_x_mean = datax[0][initial_x_mean_arg + dx]

        test_data_y = np.sum(dataz, 1)
        test_floor = np.median(test_data_y)
        test_norm = np.max(test_data_y) - test_floor

        initial_y_mean_arg = int(len(datay[:, 0]) / 2)
        model = gaussian(datay[:, 0], test_norm, test_floor, datay[:, 0][initial_y_mean_arg], initial_y_sigma)
        dy = np.argmax(np.convolve(np.sum(dataz, 1), model)) - np.argmax(np.convolve(model, model))
        initial_y_mean = datay[:, 0][initial_y_mean_arg + dy]

    else:
        initial_x_mean, initial_y_mean = np.int_(point_xy)

    if sigma is None:
        initial_x_sigma = np.abs(datax[0][np.argmin(np.abs(np.sum(dataz, 0) - np.max(np.sum(dataz, 1)) / 2))] -
                                 initial_x_mean)

        initial_y_sigma = np.abs(datay[:, 0][np.argmin(np.abs(np.sum(dataz, 1) - np.max(np.sum(dataz, 0)) / 2))] -
                                 initial_y_mean)

    if positive and symmetric:

        def gaussian_to_fit(xy_array, model_norm, model_floor, model_x_mean, model_y_mean, model_sigma):
            return two_d_gaussian(xy_array[0], xy_array[1], np.abs(model_norm), model_floor, model_x_mean,
                                  model_y_mean, model_sigma, model_sigma, 0).flatten()

        initial_values = [initial_norm, initial_floor, initial_x_mean, initial_y_mean, initial_x_sigma]

    elif positive:

        def gaussian_to_fit(xy_array, model_norm, model_floor, model_x_mean, model_y_mean, model_x_sigma, model_y_sigma,
                            model_theta):
            return two_d_gaussian(xy_array[0], xy_array[1], np.abs(model_norm), model_floor, model_x_mean,
                                  model_y_mean, model_x_sigma, model_y_sigma, model_theta).flatten()

        initial_values = [initial_norm, initial_floor, initial_x_mean, initial_y_mean,
                          initial_x_sigma, initial_y_sigma, 0]

    elif symmetric:

        def gaussian_to_fit(xy_array, model_norm, model_floor, model_x_mean, model_y_mean, model_sigma):
            return two_d_gaussian(xy_array[0], xy_array[1], model_norm, model_floor, model_x_mean,
                                  model_y_mean, model_sigma, model_sigma, 0).flatten()

        initial_values = [initial_norm, initial_floor, initial_x_mean, initial_y_mean, initial_x_sigma]

    else:

        def gaussian_to_fit(xy_array, model_norm, model_floor, model_x_mean, model_y_mean, model_x_sigma, model_y_sigma,
                            model_theta):
            return two_d_gaussian(xy_array[0], xy_array[1], model_norm, model_floor, model_x_mean, model_y_mean,
                                  model_x_sigma, model_y_sigma, model_theta).flatten()

        initial_values = [initial_norm, initial_floor, initial_x_mean, initial_y_mean,
                          initial_x_sigma, initial_y_sigma, 0]

    popt, pcov = curve_fit(gaussian_to_fit, (datax, datay), dataz.flatten(), p0=initial_values, maxfev=maxfev)

    if positive:
        popt[0] = np.abs(popt[0])
        popt[1] = np.abs(popt[1])

    popt[4] = np.abs(popt[4])
    if not symmetric:
        popt[5] = np.abs(popt[5])

    return popt, pcov




__all__ = ['EmceeFitting', 'values_to_print']

import emcee
import numpy as np
import warnings

from pylightcurve.errors import *
from pylightcurve.processes.counter import Counter
from pylightcurve.processes.files import save_dict
from pylightcurve.plots.plots_fitting import plot_mcmc_corner, plot_mcmc_traces, plot_mcmc_fitting
from pylightcurve.analysis.distributions import one_d_distribution


#  emcee


class EmceeFitting:

    def __init__(self, input_data_x, input_data_y, input_data_y_unc,
                 model, initials, limits1, limits2, walkers, iterations, burn_in,
                 data_x_name='x', data_y_name='y', data_x_print_name='x', data_y_print_name='y',
                 parameters_names=None, parameters_print_names=None, counter='MCMC', strech_prior=1000.0,
                 function_to_call=None):

        self.input_data_x = input_data_x
        self.input_data_y = input_data_y
        self.input_data_y_unc = input_data_y_unc

        self.data_x_name = data_x_name
        self.data_y_name = data_y_name
        self.data_x_print_name = data_x_print_name
        self.data_y_print_name = data_y_print_name

        self.model = model

        self.initials = np.array(initials)

        self.limits1 = np.array(limits1)

        self.limits2 = np.array(limits2)

        self.walkers = int(walkers)

        self.iterations_per_walker = int(int(iterations) / walkers)
        self.iterations = self.iterations_per_walker * walkers

        self.burn_in = int(burn_in)

        self.names = ['p{0}'.format(ff) for ff in range(len(initials))]
        if parameters_names:
            self.names = parameters_names

        self.print_names = ['p{0}'.format(ff) for ff in range(len(initials))]
        if parameters_print_names:
            self.print_names = parameters_print_names

        self.counter = Counter(counter, self.iterations + self.walkers, 100)

        self.strech_prior = strech_prior

        self.results = {
            'mcmc': {
                'iterations': self.iterations,
                'walkers': self.walkers,
                'burn_in': self.burn_in,
            },
            'input_series': {},
            'parameters': {},
            'parameters_final': [],
            'output_series': {},
            'statistics': {}}

        self.fitted_parameters = []

        self.mcmc_run_complete = False

        self.function_to_call = function_to_call

    def run_mcmc(self):

        fitted_parameters_indices = np.where(~np.isnan(self.limits1 * self.limits2))[0]

        dimensions = len(fitted_parameters_indices)

        internal_limits1 = self.limits1[fitted_parameters_indices]
        internal_limits2 = self.limits2[fitted_parameters_indices]
        internal_initials = self.initials[fitted_parameters_indices]

        walkers_initial_positions = np.random.uniform(
            (internal_initials -
             (internal_initials - internal_limits1) / self.strech_prior)[:, None] * np.ones(self.walkers),
            (internal_initials +
             (internal_limits2 - internal_initials) / self.strech_prior)[:, None] * np.ones(self.walkers))
        walkers_initial_positions = np.swapaxes(walkers_initial_positions, 0, 1)

        def internal_model(theta):
            parameters = self.initials
            parameters[fitted_parameters_indices] = theta
            return self.model(self.input_data_x, *parameters)

        def likelihood(theta):
            if ((internal_limits1 < theta) & (theta < internal_limits2)).all():
                chi = (self.input_data_y - internal_model(theta)) / self.input_data_y_unc
                return -0.5 * (np.sum(chi * chi) +
                               np.sum(np.log(2.0 * np.pi * (self.input_data_y_unc * self.input_data_y_unc))))
            else:
                return -np.inf

        def prior(theta):
            if ((internal_limits1 < theta) & (theta < internal_limits2)).all():
                return 0.0
            return -np.inf

        # probability

        def probability_core(theta):
            return prior(theta) + likelihood(theta)

        probability_core_function = probability_core
        if self.function_to_call:
            def probability_core_function(theta):
                self.function_to_call()
                return probability_core(theta)

        def probability(theta):
            self.counter.update()
            return probability_core_function(theta)

        # run sampler

        sampler = emcee.EnsembleSampler(self.walkers, dimensions, probability)
        sampler.run_mcmc(walkers_initial_positions, int(self.iterations) // int(self.walkers))
        mcmc_results = sampler.flatchain

        self.results['input_series'][self.data_x_name] = self.input_data_x
        self.results['input_series'][self.data_y_name] = self.input_data_y
        self.results['input_series']['{0}_unc'.format(self.data_y_name)] = self.input_data_y_unc
        self.results['input_series_x'] = self.data_x_name
        self.results['input_series_x_print'] = self.data_x_print_name
        self.results['input_series_y'] = self.data_y_name
        self.results['input_series_y_print'] = self.data_y_print_name
        self.results['input_series_y_unc'] = '{0}_unc'.format(self.data_y_name)

        trace_to_analyse = 0
        vars_check = 0
        for var in range(len(self.names)):

            if not np.isnan(self.limits1[var]):

                trace = mcmc_results[:, np.where(fitted_parameters_indices == var)[0][0]]
                trace = trace.reshape(int(self.walkers), int(self.iterations) // int(self.walkers))
                trace = (np.swapaxes(trace, 0, 1).flatten())[self.burn_in:]

                median = np.median(trace)
                mad = np.sqrt(np.median((trace - median) ** 2))

                trace_to_analyse += (trace > (median - 5 * mad)) * (trace < (median + 5 * mad))

                vars_check += 1

        trace_to_analyse = np.where(trace_to_analyse == vars_check)

        for var in range(len(self.names)):

            if np.isnan(self.limits1[var]):

                variable = {'name': self.names[var], 'print_name': self.print_names[var],
                            'initial': None, 'min_allowed': None, 'max_allowed': None,
                            'trace': None, 'trace_bins': None, 'trace_counts': None,
                            'value': self.initials[var], 'm_error': None, 'p_error': None,
                            'print_value': self.initials[var], 'print_m_error': '-', 'print_p_error': '-'}

            else:
                trace = mcmc_results[:, np.where(fitted_parameters_indices == var)[0][0]]
                trace = trace.reshape(int(self.walkers), int(self.iterations) // int(self.walkers))
                trace = (np.swapaxes(trace, 0, 1).flatten())[self.burn_in:]

                trace = trace[trace_to_analyse]

                bins, counts, value, m_error, p_error = \
                    one_d_distribution(trace, confidence_interval=0.68)

                print_value, print_m_error, print_p_error = values_to_print(value, m_error, p_error)

                variable = {'name': self.names[var], 'print_name': self.print_names[var],
                            'initial': self.initials[var],
                            'min_allowed': self.limits1[var], 'max_allowed': self.limits2[var],
                            'trace': trace, 'trace_bins': bins, 'trace_counts': counts,
                            'value': value, 'm_error': m_error, 'p_error': p_error,
                            'print_value': print_value, 'print_m_error': print_m_error, 'print_p_error': print_p_error}

                self.fitted_parameters.append(self.names[var])

            self.results['parameters'][self.names[var]] = variable
            self.results['parameters_final'].append(variable['value'])

        self.results['output_series']['model'] = self.model(self.input_data_x, *self.results['parameters_final'])
        self.results['output_series']['residuals'] = self.input_data_y - self.results['output_series']['model']

        to_correlate = []
        for parameter in self.fitted_parameters:
            to_correlate.append(self.results['parameters'][parameter]['trace'])
        correlation_matrix = np.corrcoef(to_correlate)

        res_autocorr = np.correlate(self.results['output_series']['residuals'],
                                    self.results['output_series']['residuals'], mode='full')
        res_autocorr = res_autocorr[res_autocorr.size // 2:] / res_autocorr[res_autocorr.size // 2:][0]

        self.results['statistics']['res_autocorr'] = res_autocorr
        self.results['statistics']['res_max_autocorr'] = np.max(res_autocorr[1:])
        self.results['statistics']['res_mean'] = np.mean(self.results['output_series']['residuals'])
        self.results['statistics']['res_std'] = np.std(self.results['output_series']['residuals'])
        self.results['statistics']['res_rms'] = np.sqrt(np.mean(self.results['output_series']['residuals']**2))
        self.results['statistics']['res_chi_sqr'] = np.sum(
            (self.results['output_series']['residuals'] ** 2) / (self.input_data_y_unc ** 2))
        self.results['statistics']['res_red_chi_sqr'] = (
                self.results['statistics']['res_chi_sqr'] / (len(self.input_data_y_unc) - len(self.fitted_parameters)))
        self.results['statistics']['corr_matrix'] = correlation_matrix
        self.results['statistics']['corr_variables'] = ','.join(self.fitted_parameters)

        self.mcmc_run_complete = True

    def save_all(self, export_file):

        if not self.mcmc_run_complete:
            raise PyLCProcessError('MCMC not completed')

        save_dict(self.results, export_file)

    def save_results(self, export_file):

        if not self.mcmc_run_complete:
            raise PyLCProcessError('MCMC not completed')

        w = open(export_file, 'w')

        w.write('# variable\tresult\tuncertainty\n')

        for i in self.names:
            w.write('{0}\t{1}\t-{2} +{3}\n'.format(self.results['parameters'][i]['name'],
                                                   self.results['parameters'][i]['print_value'],
                                                   self.results['parameters'][i]['print_m_error'],
                                                   self.results['parameters'][i]['print_p_error']))

        w.write('\n#Residuals:\n')
        w.write('#Mean: {0}\n'.format(self.results['statistics']['res_mean']))
        w.write('#STD: {0}\n'.format(self.results['statistics']['res_std']))
        w.write('#RMS: {0}\n'.format(self.results['statistics']['res_rms']))
        w.write('#Max auto-correlation: {0}\n'.format(self.results['statistics']['res_max_autocorr']))
        w.write('#Chi squared: {0}\n'.format(self.results['statistics']['res_chi_sqr']))
        w.write('#Reduced chi squared: {0}\n'.format(self.results['statistics']['res_red_chi_sqr']))

        w.close()

    def plot_fitting(self, export_file):

        plot_mcmc_fitting(self, export_file)

    def plot_corner(self, export_file):

        plot_mcmc_corner(self, export_file)

    def plot_traces(self, export_file):

        plot_mcmc_traces(self, export_file)


# decimal points and rounding


def values_to_print(value, error_minus, error_plus):

    value = float(value)
    error_minus = float(error_minus)
    error_plus = float(error_plus)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        if error_minus >= 1.0 or error_minus == 0.0:
            digit1 = 1
        else:
            str_error_minus = '{0:.{test}f}'.format(error_minus, test=10 + abs(int(np.log10(error_minus))))
            digit1 = np.where([ff not in ['0', '.'] for ff in str_error_minus])[0][0] - 1

        if error_plus >= 1.0 or error_plus == 0.0:
            digit2 = 1
        else:
            str_error_plus = '{0:.{test}f}'.format(error_plus, test=10 + abs(int(np.log10(error_plus))))
            digit2 = np.where([ff not in ['0', '.'] for ff in str_error_plus])[0][0] - 1

    width = max(1, digit1, digit2)

    print_m_error = '{0:.{width}f}'.format(round(error_minus, width), width=width)
    print_p_error = '{0:.{width}f}'.format(round(error_plus, width), width=width)

    if print_m_error[-1] in ['1', '2'] and float(print_m_error[:-1]) == 0:
        width += 1
    elif print_p_error[-1] in ['1', '2'] and float(print_p_error[:-1]) == 0:
        width += 1

    if error_plus >= 1.0 and error_minus >= 1.0:
        width = 1

    print_value = '{0:.{width}f}'.format(round(value, width), width=width)
    print_m_error = '{0:.{width}f}'.format(round(error_minus, width), width=width)
    print_p_error = '{0:.{width}f}'.format(round(error_plus, width), width=width)

    return print_value, print_m_error, print_p_error




__all__ = ['planet_orbit', 'planet_star_projected_distance', 'planet_phase',
           'transit', 'transit_integrated', 'transit_depth', 'transit_duration',
           'eclipse', 'eclipse_integrated', 'eclipse_depth', 'eclipse_duration', 'eclipse_mid_time',
           'fp_over_fs', 'exotethys']


import os
import numpy as np

from pylightcurve.errors import *
from pylightcurve.__databases__ import plc_data
from pylightcurve.analysis.numerical_integration import gauss_numerical_integration
from pylightcurve.analysis.curve_fit import curve_fit
from pylightcurve.processes.files import open_dict


# orbit


def planet_orbit(period, sma_over_rs, eccentricity, inclination, periastron, mid_time, time_array, ww=0):

    inclination = inclination * np.pi / 180.0
    periastron = periastron * np.pi / 180.0
    ww = ww * np.pi / 180.0

    if eccentricity == 0 and ww == 0:
        vv = 2 * np.pi * (time_array - mid_time) / period
        bb = sma_over_rs * np.cos(vv)
        return [bb * np.sin(inclination), sma_over_rs * np.sin(vv), - bb * np.cos(inclination)]

    if periastron < np.pi / 2:
        aa = 1.0 * np.pi / 2 - periastron
    else:
        aa = 5.0 * np.pi / 2 - periastron
    bb = 2 * np.arctan(np.sqrt((1 - eccentricity) / (1 + eccentricity)) * np.tan(aa / 2))
    if bb < 0:
        bb += 2 * np.pi
    mid_time = float(mid_time) - (period / 2.0 / np.pi) * (bb - eccentricity * np.sin(bb))
    m = (time_array - mid_time - np.int_((time_array - mid_time) / period) * period) * 2.0 * np.pi / period
    u0 = m
    stop = False
    u1 = 0
    for ii in range(10000):  # setting a limit of 1k iterations - arbitrary limit
        u1 = u0 - (u0 - eccentricity * np.sin(u0) - m) / (1 - eccentricity * np.cos(u0))
        stop = (np.abs(u1 - u0) < 10 ** (-7)).all()
        if stop:
            break
        else:
            u0 = u1
    if not stop:
        raise RuntimeError('Failed to find a solution in 10000 loops')

    vv = 2 * np.arctan(np.sqrt((1 + eccentricity) / (1 - eccentricity)) * np.tan(u1 / 2))
    #
    rr = sma_over_rs * (1 - (eccentricity ** 2)) / (np.ones_like(vv) + eccentricity * np.cos(vv))
    aa = np.cos(vv + periastron)
    bb = np.sin(vv + periastron)
    x = rr * bb * np.sin(inclination)
    y = rr * (-aa * np.cos(ww) + bb * np.sin(ww) * np.cos(inclination))
    z = rr * (-aa * np.sin(ww) - bb * np.cos(ww) * np.cos(inclination))

    return [x, y, z]


def planet_star_projected_distance(period, sma_over_rs, eccentricity, inclination, periastron, mid_time, time_array):

    position_vector = planet_orbit(period, sma_over_rs, eccentricity, inclination, periastron, mid_time, time_array)

    return np.sqrt(position_vector[1] * position_vector[1] + position_vector[2] * position_vector[2])


def planet_phase(period, mid_time, time_array):
    return (time_array - mid_time)/period


# flux drop


def integral_r_claret(limb_darkening_coefficients, r):
    a1, a2, a3, a4 = limb_darkening_coefficients
    mu44 = 1.0 - r * r
    mu24 = np.sqrt(mu44)
    mu14 = np.sqrt(mu24)
    return - (2.0 * (1.0 - a1 - a2 - a3 - a4) / 4) * mu44 \
           - (2.0 * a1 / 5) * mu44 * mu14 \
           - (2.0 * a2 / 6) * mu44 * mu24 \
           - (2.0 * a3 / 7) * mu44 * mu24 * mu14 \
           - (2.0 * a4 / 8) * mu44 * mu44


def num_claret(r, limb_darkening_coefficients, rprs, z):
    a1, a2, a3, a4 = limb_darkening_coefficients
    rsq = r * r
    mu44 = 1.0 - rsq
    mu24 = np.sqrt(mu44)
    mu14 = np.sqrt(mu24)
    return ((1.0 - a1 - a2 - a3 - a4) + a1 * mu14 + a2 * mu24 + a3 * mu24 * mu14 + a4 * mu44) \
        * r * np.arccos(np.minimum((-rprs ** 2 + z * z + rsq) / (2.0 * z * r), 1.0))


def integral_r_f_claret(limb_darkening_coefficients, rprs, z, r1, r2, precision=3):
    return gauss_numerical_integration(num_claret, r1, r2, precision, limb_darkening_coefficients, rprs, z)


# integral definitions for zero method


def integral_r_zero(limb_darkening_coefficients, r):
    musq = 1 - r * r
    return (-1.0 / 6) * musq * 3.0


def num_zero(r, limb_darkening_coefficients, rprs, z):
    rsq = r * r
    return r * np.arccos(np.minimum((-rprs ** 2 + z * z + rsq) / (2.0 * z * r), 1.0))


def integral_r_f_zero(limb_darkening_coefficients, rprs, z, r1, r2, precision=3):
    return gauss_numerical_integration(num_zero, r1, r2, precision, limb_darkening_coefficients, rprs, z)


# integral definitions for linear method


def integral_r_linear(limb_darkening_coefficients, r):
    a1 = limb_darkening_coefficients[0]
    musq = 1 - r * r
    return (-1.0 / 6) * musq * (3.0 + a1 * (-3.0 + 2.0 * np.sqrt(musq)))


def num_linear(r, limb_darkening_coefficients, rprs, z):
    a1 = limb_darkening_coefficients[0]
    rsq = r * r
    return (1.0 - a1 * (1.0 - np.sqrt(1.0 - rsq))) \
        * r * np.arccos(np.minimum((-rprs ** 2 + z * z + rsq) / (2.0 * z * r), 1.0))


def integral_r_f_linear(limb_darkening_coefficients, rprs, z, r1, r2, precision=3):
    return gauss_numerical_integration(num_linear, r1, r2, precision, limb_darkening_coefficients, rprs, z)


# integral definitions for quadratic method


def integral_r_quad(limb_darkening_coefficients, r):
    a1, a2 = limb_darkening_coefficients[:2]
    musq = 1 - r * r
    mu = np.sqrt(musq)
    return (1.0 / 12) * (-4.0 * (a1 + 2.0 * a2) * mu * musq + 6.0 * (-1 + a1 + a2) * musq + 3.0 * a2 * musq * musq)


def num_quad(r, limb_darkening_coefficients, rprs, z):
    a1, a2 = limb_darkening_coefficients[:2]
    rsq = r * r
    cc = 1.0 - np.sqrt(1.0 - rsq)
    return (1.0 - a1 * cc - a2 * cc * cc) \
        * r * np.arccos(np.minimum((-rprs ** 2 + z * z + rsq) / (2.0 * z * r), 1.0))


def integral_r_f_quad(limb_darkening_coefficients, rprs, z, r1, r2, precision=3):
    return gauss_numerical_integration(num_quad, r1, r2, precision, limb_darkening_coefficients, rprs, z)


# integral definitions for square root method


def integral_r_sqrt(limb_darkening_coefficients, r):
    a1, a2 = limb_darkening_coefficients[:2]
    musq = 1 - r * r
    mu = np.sqrt(musq)
    return ((-2.0 / 5) * a2 * np.sqrt(mu) - (1.0 / 3) * a1 * mu + (1.0 / 2) * (-1 + a1 + a2)) * musq


def num_sqrt(r, limb_darkening_coefficients, rprs, z):
    a1, a2 = limb_darkening_coefficients[:2]
    rsq = r * r
    mu = np.sqrt(1.0 - rsq)
    return (1.0 - a1 * (1 - mu) - a2 * (1.0 - np.sqrt(mu))) \
        * r * np.arccos(np.minimum((-rprs ** 2 + z * z + rsq) / (2.0 * z * r), 1.0))


def integral_r_f_sqrt(limb_darkening_coefficients, rprs, z, r1, r2, precision=3):
    return gauss_numerical_integration(num_sqrt, r1, r2, precision, limb_darkening_coefficients, rprs, z)


# dictionaries containing the different methods,
# if you define a new method, include the functions in the dictionary as well

integral_r = {
    'claret': integral_r_claret,
    'linear': integral_r_linear,
    'quad': integral_r_quad,
    'sqrt': integral_r_sqrt,
    'zero': integral_r_zero
}

integral_r_f = {
    'claret': integral_r_f_claret,
    'linear': integral_r_f_linear,
    'quad': integral_r_f_quad,
    'sqrt': integral_r_f_sqrt,
    'zero': integral_r_f_zero,
}


def integral_centred(method, limb_darkening_coefficients, rprs, ww1, ww2):
    return (integral_r[method](limb_darkening_coefficients, rprs)
            - integral_r[method](limb_darkening_coefficients, 0.0)) * np.abs(ww2 - ww1)


def integral_plus_core(method, limb_darkening_coefficients, rprs, z, ww1, ww2, precision=3):
    if len(z) == 0:
        return z
    rr1 = z * np.cos(ww1) + np.sqrt(np.maximum(rprs ** 2 - (z * np.sin(ww1)) ** 2, 0))
    rr1 = np.clip(rr1, 0, 1)
    rr2 = z * np.cos(ww2) + np.sqrt(np.maximum(rprs ** 2 - (z * np.sin(ww2)) ** 2, 0))
    rr2 = np.clip(rr2, 0, 1)
    w1 = np.minimum(ww1, ww2)
    r1 = np.minimum(rr1, rr2)
    w2 = np.maximum(ww1, ww2)
    r2 = np.maximum(rr1, rr2)
    parta = integral_r[method](limb_darkening_coefficients, 0.0) * (w1 - w2)
    partb = integral_r[method](limb_darkening_coefficients, r1) * w2
    partc = integral_r[method](limb_darkening_coefficients, r2) * (-w1)
    partd = integral_r_f[method](limb_darkening_coefficients, rprs, z, r1, r2, precision=precision)
    return parta + partb + partc + partd


def integral_minus_core(method, limb_darkening_coefficients, rprs, z, ww1, ww2, precision=3):
    if len(z) == 0:
        return z
    rr1 = z * np.cos(ww1) - np.sqrt(np.maximum(rprs ** 2 - (z * np.sin(ww1)) ** 2, 0))
    rr1 = np.clip(rr1, 0, 1)
    rr2 = z * np.cos(ww2) - np.sqrt(np.maximum(rprs ** 2 - (z * np.sin(ww2)) ** 2, 0))
    rr2 = np.clip(rr2, 0, 1)
    w1 = np.minimum(ww1, ww2)
    r1 = np.minimum(rr1, rr2)
    w2 = np.maximum(ww1, ww2)
    r2 = np.maximum(rr1, rr2)
    parta = integral_r[method](limb_darkening_coefficients, 0.0) * (w1 - w2)
    partb = integral_r[method](limb_darkening_coefficients, r1) * (-w1)
    partc = integral_r[method](limb_darkening_coefficients, r2) * w2
    partd = integral_r_f[method](limb_darkening_coefficients, rprs, z, r1, r2, precision=precision)
    return parta + partb + partc - partd


def transit_flux_drop(limb_darkening_coefficients, rp_over_rs, z_over_rs, method='claret', precision=3):

    z_over_rs = np.where(z_over_rs < 0, 1.0 + 100.0 * rp_over_rs, z_over_rs)
    z_over_rs = np.maximum(z_over_rs, 10**(-10))

    # cases
    zsq = z_over_rs * z_over_rs
    sum_z_rprs = z_over_rs + rp_over_rs
    dif_z_rprs = rp_over_rs - z_over_rs
    sqr_dif_z_rprs = zsq - rp_over_rs ** 2
    case0 = np.where((z_over_rs == 0) & (rp_over_rs <= 1))
    case1 = np.where((z_over_rs < rp_over_rs) & (sum_z_rprs <= 1))
    casea = np.where((z_over_rs < rp_over_rs) & (sum_z_rprs > 1) & (dif_z_rprs < 1))
    caseb = np.where((z_over_rs < rp_over_rs) & (sum_z_rprs > 1) & (dif_z_rprs > 1))
    case2 = np.where((z_over_rs == rp_over_rs) & (sum_z_rprs <= 1))
    casec = np.where((z_over_rs == rp_over_rs) & (sum_z_rprs > 1))
    case3 = np.where((z_over_rs > rp_over_rs) & (sum_z_rprs < 1))
    case4 = np.where((z_over_rs > rp_over_rs) & (sum_z_rprs == 1))
    case5 = np.where((z_over_rs > rp_over_rs) & (sum_z_rprs > 1) & (sqr_dif_z_rprs < 1))
    case6 = np.where((z_over_rs > rp_over_rs) & (sum_z_rprs > 1) & (sqr_dif_z_rprs == 1))
    case7 = np.where((z_over_rs > rp_over_rs) & (sum_z_rprs > 1) & (sqr_dif_z_rprs > 1) & (-1 < dif_z_rprs))
    plus_case = np.concatenate((case1[0], case2[0], case3[0], case4[0], case5[0], casea[0], casec[0]))
    minus_case = np.concatenate((case3[0], case4[0], case5[0], case6[0], case7[0]))
    star_case = np.concatenate((case5[0], case6[0], case7[0], casea[0], casec[0]))

    # cross points
    ph = np.arccos(np.clip((1.0 - rp_over_rs ** 2 + zsq) / (2.0 * z_over_rs), -1, 1))
    theta_1 = np.zeros(len(z_over_rs))
    ph_case = np.concatenate((case5[0], casea[0], casec[0]))
    theta_1[ph_case] = ph[ph_case]
    theta_2 = np.arcsin(np.minimum(rp_over_rs / z_over_rs, 1))
    theta_2[case1] = np.pi
    theta_2[case2] = np.pi / 2.0
    theta_2[casea] = np.pi
    theta_2[casec] = np.pi / 2.0
    theta_2[case7] = ph[case7]

    # flux_upper
    plusflux = np.zeros(len(z_over_rs))
    plusflux[plus_case] = integral_plus_core(method, limb_darkening_coefficients, rp_over_rs, z_over_rs[plus_case],
                                             theta_1[plus_case], theta_2[plus_case], precision=precision)
    if len(case0[0]) > 0:
        plusflux[case0] = integral_centred(method, limb_darkening_coefficients, rp_over_rs, 0.0, np.pi)
    if len(caseb[0]) > 0:
        plusflux[caseb] = integral_centred(method, limb_darkening_coefficients, 1, 0.0, np.pi)

    # flux_lower
    minsflux = np.zeros(len(z_over_rs))
    minsflux[minus_case] = integral_minus_core(method, limb_darkening_coefficients, rp_over_rs,
                                               z_over_rs[minus_case], 0.0, theta_2[minus_case], precision=precision)

    # flux_star
    starflux = np.zeros(len(z_over_rs))
    starflux[star_case] = integral_centred(method, limb_darkening_coefficients, 1, 0.0, ph[star_case])

    # flux_total
    total_flux = integral_centred(method, limb_darkening_coefficients, 1, 0.0, 2.0 * np.pi)

    return 1 - (2.0 / total_flux) * (plusflux + starflux - minsflux)


# transit

def transit(limb_darkening_coefficients, rp_over_rs, period, sma_over_rs, eccentricity, inclination, periastron,
            mid_time, time_array, method='claret', precision=3):

    position_vector = planet_orbit(period, sma_over_rs, eccentricity, inclination, periastron, mid_time, time_array)

    projected_distance = np.where(
        position_vector[0] < 0, 1.0 + 5.0 * rp_over_rs,
        np.sqrt(position_vector[1] * position_vector[1] + position_vector[2] * position_vector[2]))

    return transit_flux_drop(limb_darkening_coefficients, rp_over_rs, projected_distance,
                             method=method, precision=precision)


def transit_integrated(limb_darkening_coefficients, rp_over_rs, period, sma_over_rs, eccentricity, inclination,
                       periastron, mid_time, time_array, exp_time, max_sub_exp_time=10, method='claret', precision=3):

    time_factor = int(exp_time / max_sub_exp_time) + 1
    exp_time /= (60.0 * 60.0 * 24.0)

    time_array_hr = (time_array[:, None] + np.arange(-exp_time / 2 + exp_time / time_factor / 2, exp_time / 2,
                                                     exp_time / time_factor)).flatten()

    position_vector = planet_orbit(period, sma_over_rs, eccentricity,
                                   inclination, periastron, mid_time, time_array_hr)

    projected_distance = np.where(
        position_vector[0] < 0, 1.0 + 5.0 * rp_over_rs,
        np.sqrt(position_vector[1] * position_vector[1] + position_vector[2] * position_vector[2]))

    return np.mean(np.reshape(
        transit_flux_drop(limb_darkening_coefficients, rp_over_rs, projected_distance,
                          method=method, precision=precision),
        (len(time_array), time_factor)), 1)


def transit_duration(rp_over_rs, period, sma_over_rs, eccentricity, inclination, periastron):

    ww = periastron * np.pi / 180
    ii = inclination * np.pi / 180
    ee = eccentricity
    aa = sma_over_rs
    ro_pt = (1 - ee ** 2) / (1 + ee * np.sin(ww))
    b_pt = aa * ro_pt * np.cos(ii)
    if b_pt > 1:
        b_pt = 0.5
    s_ps = 1.0 + rp_over_rs
    df = np.arcsin(np.sqrt((s_ps ** 2 - b_pt ** 2) / ((aa ** 2) * (ro_pt ** 2) - b_pt ** 2)))
    aprox = (period * (ro_pt ** 2)) / (np.pi * np.sqrt(1 - ee ** 2)) * df * 60 * 60 * 24

    def function_to_fit(x, t):
        return planet_star_projected_distance(period, sma_over_rs, eccentricity, inclination, periastron,
                                              10000, np.array(10000 + t / 24 / 60 / 60))

    popt1, pcov1 = curve_fit(function_to_fit, [0], [1.0 + rp_over_rs], p0=[-aprox / 2])
    popt2, pcov2 = curve_fit(function_to_fit, [0], [1.0 + rp_over_rs], p0=[aprox / 2])

    return (popt2[0] - popt1[0]) / 24 / 60 / 60


def transit_depth(limb_darkening_coefficients, rp_over_rs, period, sma_over_rs, eccentricity, inclination,
                  periastron, method='claret', precision=6):

    return 1 - transit(limb_darkening_coefficients, rp_over_rs, period, sma_over_rs, eccentricity, inclination,
                       periastron, 10000, np.array([10000]), method=method, precision=precision)[0]


# eclipse

def eclipse(fp_over_fs, rp_over_rs, period, sma_over_rs, eccentricity, inclination, periastron, mid_time,
            time_array, precision=3):

    position_vector = planet_orbit(period, sma_over_rs / rp_over_rs, eccentricity, inclination, periastron + 180,
                                   mid_time, time_array)

    projected_distance = np.where(
        position_vector[0] < 0, 1.0 + 5.0 / rp_over_rs,
        np.sqrt(position_vector[1] * position_vector[1] + position_vector[2] * position_vector[2]))

    return (1.0 + fp_over_fs * transit_flux_drop([0, 0, 0, 0], 1 / rp_over_rs, projected_distance,
                                                 precision=precision, method='zero')) / (1.0 + fp_over_fs)


def eclipse_integrated(fp_over_fs, rp_over_rs, period, sma_over_rs, eccentricity, inclination, periastron,
                       mid_time, time_array, exp_time, max_sub_exp_time=10, precision=3):

    time_factor = int(exp_time / max_sub_exp_time) + 1
    exp_time /= (60.0 * 60.0 * 24.0)

    time_array_hr = (time_array[:, None] + np.arange(-exp_time / 2 + exp_time / time_factor / 2, exp_time / 2,
                                                     exp_time / time_factor)).flatten()

    position_vector = planet_orbit(period, sma_over_rs / rp_over_rs, eccentricity, inclination, periastron + 180,
                                   mid_time, time_array_hr)

    projected_distance = np.where(
        position_vector[0] < 0, 1.0 + 5.0 * rp_over_rs,
        np.sqrt(position_vector[1] * position_vector[1] + position_vector[2] * position_vector[2]))

    return np.mean(np.reshape(
        (1.0 + fp_over_fs * transit_flux_drop([0, 0, 0, 0], 1 / rp_over_rs, projected_distance, method='zero',
                                              precision=precision)) / (1.0 + fp_over_fs),
        (len(time_array), time_factor)), 1)


def eclipse_mid_time(period, sma_over_rs, eccentricity, inclination, periastron, mid_time):
    test_array = np.arange(0, period, 0.001)
    xx, yy, zz = planet_orbit(period, sma_over_rs, eccentricity, inclination, periastron, mid_time,
                              test_array + mid_time)

    test1 = np.where(xx < 0)
    yy = yy[test1]
    test_array = test_array[test1]

    aprox = test_array[np.argmin(np.abs(yy))]

    def function_to_fit(x, t):
        xx, yy, zz = planet_orbit(period, sma_over_rs, eccentricity, inclination, periastron, mid_time,
                            np.array(mid_time + t))
        return yy

    popt, pcov = curve_fit(function_to_fit, [0], [0], p0=[aprox])

    return mid_time + popt[0]


def eclipse_duration(rp_over_rs, period, sma_over_rs, eccentricity, inclination, periastron):

    ww = periastron * np.pi / 180
    ii = inclination * np.pi / 180
    ee = eccentricity
    aa = sma_over_rs
    ro_pt = (1 - ee ** 2) / (1 + ee * np.sin(ww))
    b_pt = aa * ro_pt * np.cos(ii)
    if b_pt > 1:
        b_pt = 0.5
    s_ps = 1.0 + rp_over_rs
    df = np.arcsin(np.sqrt((s_ps ** 2 - b_pt ** 2) / ((aa ** 2) * (ro_pt ** 2) - b_pt ** 2)))
    aprox = (period * (ro_pt ** 2)) / (np.pi * np.sqrt(1 - ee ** 2)) * df * 60 * 60 * 24

    emt = eclipse_mid_time(period, sma_over_rs, eccentricity, inclination, periastron, 10000)

    def function_to_fit(x, t):
        xx = planet_star_projected_distance(period, sma_over_rs, eccentricity, inclination, periastron,
                                              10000, np.array(emt + t / 24 / 60 / 60))
        return xx

    popt1, pcov1 = curve_fit(function_to_fit, [0], [1.0 + rp_over_rs], p0=[-aprox / 2])
    popt2, pcov2 = curve_fit(function_to_fit, [0], [1.0 + rp_over_rs], p0=[aprox / 2])

    return (popt2[0] - popt1[0]) / 24 / 60 / 60


def eclipse_depth(fp_over_fs, rp_over_rs, period, sma_over_rs, eccentricity, inclination, periastron, precision=6):

    return 1 - eclipse(fp_over_fs, rp_over_rs, period, sma_over_rs, eccentricity, inclination,
                       periastron, 10000, np.array([10000]), precision=precision)[0]


def _get_filter(photometric_filter):

    if photometric_filter not in plc_data.all_filters():
        raise PyLCInputError('{0} is not available. Available filters: {1}'.format(
            photometric_filter, ','.join(plc_data.all_filters())))


def fp_over_fs(rp_over_rs, sma_over_rs, albedo, emissivity, stellar_temperature, filter_name):

    _get_filter(filter_name)

    def _black_body(w, t):
        # w in mu
        w = w / (10 ** 6)
        h = 6.62607004 * (10 ** (-34))
        c = 3 * (10 ** 8)
        w5 = w ** 5
        k = 1.38064852 * (10 ** (-23))
        return (2 * h * c * c / w5) / (np.exp(h * c / w / k / t) - 1)

    planet_temperature = stellar_temperature * np.sqrt(0.5 / sma_over_rs) * (((1 - albedo) / emissivity) ** 0.25)

    wavelength_array, band = np.loadtxt(os.path.join(plc_data.photometry(), filter_name + '.pass'), unpack=True)
    binsedge = 0.5 * (wavelength_array[:-1] + wavelength_array[1:])
    binsedge1 = np.append(wavelength_array[0] - (binsedge[0] - wavelength_array[0]), binsedge)
    binsedge2 = np.append(binsedge, wavelength_array[-1] + (wavelength_array[-1] - binsedge[-1]))
    binswidth = binsedge2 - binsedge1

    weights = band * binswidth / 10000
    emission = ((rp_over_rs ** 2) *
                _black_body(wavelength_array / 10000, planet_temperature) /
                _black_body(wavelength_array / 10000, stellar_temperature))

    emission = np.sum(emission * weights) / np.sum(weights)

    reflection = albedo * (rp_over_rs ** 2) / (sma_over_rs ** 2)

    return reflection + emission


def exotethys(stellar_logg, stellar_temperature, stellar_metallicity, filter_name, method='claret',
              stellar_model='phoenix'):

    available_methods = ['claret', 'sqrt', 'quad', 'linear']

    if not isinstance(method, str) or method.lower() not in available_methods:
        raise PyLCInputError('{0} is not available. Available methods: {1}'.format(
            method, ','.join(available_methods)))

    method = method.lower()

    method_map = {
        'claret': 'claret4',
        'sqrt': 'square_root',
        'quad': 'quadratic',
        'linear': 'linear'
    }

    _get_filter(filter_name)

    available_stellar_models = ['atlas', 'phoenix']

    if not isinstance(stellar_model, str) or stellar_model.lower() not in available_stellar_models:
        raise PyLCInputError('{0} is not available. Available stellar models: {1}'.format(
            stellar_model, ','.join(available_stellar_models)))

    stellar_model = stellar_model.lower()

    if stellar_model == 'phoenix' and stellar_metallicity != 0:
        print('PHOENIX models are only computed for solar metallicity stars. Setting stellar_metallicity = 0.')
        stellar_metallicity = 0
    elif stellar_model == 'phoenix' and filter_name in ['irac1', 'irac2', 'irac3', 'irac4']:
        print('PHOENIX models are not available for this filter. Setting stellar_model = atlas')
        stellar_model = 'atlas'

    data = open_dict(os.path.join(plc_data.exotethys(), '{0}_{1}.pickle'.format(stellar_model, filter_name)))

    xin = float(stellar_logg)
    yin = float(stellar_temperature)
    zin = float(stellar_metallicity)

    xunique = np.unique(data['logg'])
    xunique = sorted(xunique, key=lambda x:(x - xin)**2)

    yunique = np.unique(data['teff'])
    yunique = sorted(yunique, key=lambda x:(x - yin)**2)

    zunique = np.unique(data['meta'])
    zunique = sorted(zunique, key=lambda x:(x - zin)**2)

    xmin = 0
    xmax = 0
    ymin = 0
    ymax = 0
    zmin = 0
    zmax = 0

    complete = False
    tsearch = 0

    while not complete and tsearch < len(yunique):

        ymin = min(yunique[:tsearch + 2])
        ymax = max(yunique[:tsearch + 2])

        tsearch += 1

        gsearch = 0

        while not complete and gsearch < len(xunique):

            xmin = min(xunique[:gsearch + 2])
            xmax = max(xunique[:gsearch + 2])

            gsearch += 1

            msearch = 0

            while not complete and msearch < len(zunique):

                zmin = min(zunique[:msearch + 2])
                zmax = max(zunique[:msearch + 2])

                msearch += 1

                # print(xmin, xmax, ymin, ymax, zmin, zmax)

                if all([
                    len(np.where((data['logg'] == ff[0]) & (data['teff'] == ff[1]) & (data['meta'] == ff[2]))[0]) > 0
                        for ff in [
                        [xmin, ymin, zmin],
                        [xmin, ymax, zmin],
                        [xmax, ymin, zmin],
                        [xmax, ymax, zmin],
                        [xmin, ymin, zmax],
                        [xmin, ymax, zmax],
                        [xmax, ymin, zmax],
                        [xmax, ymax, zmax],

                    ]]):

                    if xmin <= xin and xmax >= xin and ymin <= yin and ymax >= yin and zmin <= zin and zmax >= zin:

                        complete = True

    if not complete:
        raise PyLCProcessError('LDCs could not be estimated for the given set of stellar parameters: '
                               'Log(g): {0} to {1} (cgs), Teff: {2} to {3} K, MH: {4} to {5} (dex)'.format(
            xmin, xmax, ymin, ymax, zmin, zmax))
    # else:
    #     print('Trilinear interpolation between: Log(g): {0} to {1} (cgs), Teff: {2} to {3} K, MH: {4} to {5} (dex)'.format(
    #         xmin, xmax, ymin, ymax, zmin, zmax))

    def tri_linear(x, y, z, x0, x1, y0, y1, z0, z1, v000, v100, v010, v001, v101, v011, v110, v111):
        c0 = v000
        c1 = v100 - v000
        c2 = v010 - v000
        c3 = v001 - v000
        c4 = v110 - v010 - v100 + v000
        c5 = v011 - v001 - v010 + v000
        c6 = v101 - v001 - v100 + v000
        c7 = v111 - v011 - v101 - v110 + v100 + v001 + v010 - v000
        if x == x0 == x1:
            dx = 0
        else:
            dx = (x - x0) / (x1 - x0)
        if y == y0 == y1:
            dy = 0
        else:
            dy = (y - y0) / (y1 - y0)
        if z == z0 == z1:
            dz = 0
        else:
            dz = (z - z0) / (z1 - z0)
        return c0 + c1 * dx + c2 * dy + c3 * dz + c4 * dx * dy + c5 * dy * dz + c6 * dz * dx + c7 * dx * dy * dz

    final_coefficients = []

    for index in [1, 2, 3, 4]:

        vv000 = data[method_map[method]]['ldc' + str(index)][np.where((data['logg'] == xmin) & (data['teff'] == ymin) & (data['meta'] == zmin))][0]
        vv100 = data[method_map[method]]['ldc' + str(index)][np.where((data['logg'] == xmax) & (data['teff'] == ymin) & (data['meta'] == zmin))][0]
        vv010 = data[method_map[method]]['ldc' + str(index)][np.where((data['logg'] == xmin) & (data['teff'] == ymax) & (data['meta'] == zmin))][0]
        vv001 = data[method_map[method]]['ldc' + str(index)][np.where((data['logg'] == xmin) & (data['teff'] == ymin) & (data['meta'] == zmax))][0]
        vv101 = data[method_map[method]]['ldc' + str(index)][np.where((data['logg'] == xmax) & (data['teff'] == ymin) & (data['meta'] == zmax))][0]
        vv011 = data[method_map[method]]['ldc' + str(index)][np.where((data['logg'] == xmin) & (data['teff'] == ymax) & (data['meta'] == zmax))][0]
        vv110 = data[method_map[method]]['ldc' + str(index)][np.where((data['logg'] == xmax) & (data['teff'] == ymax) & (data['meta'] == zmin))][0]
        vv111 = data[method_map[method]]['ldc' + str(index)][np.where((data['logg'] == xmax) & (data['teff'] == ymax) & (data['meta'] == zmax))][0]

        res = tri_linear(xin, yin, zin, xmin, xmax, ymin, ymax, zmin, zmax,
                         vv000, vv100, vv010, vv001, vv101, vv011, vv110, vv111)

        final_coefficients.append(res)

    return np.array(final_coefficients)




__all__ = ['one_d_distribution', 'two_d_distribution']

import numpy as np

from pylightcurve.errors import *
from pylightcurve.analysis.gaussian import fit_gaussian, fit_two_d_gaussian


def one_d_distribution(datax, step=None, abs_step=None, min_value=None, max_value=None, mad_filter=None,
                       confidence_interval=None, gaussian_fit=False):

    datax = 1.0 * datax.flatten()

    if min_value is None and max_value is None:
        if mad_filter:
            data_median = np.median(datax)
            data_mad = np.sqrt(np.median((datax - data_median) ** 2))
            min_value = data_median - mad_filter * data_mad
            max_value = data_median + mad_filter * data_mad
        else:
            min_value = np.min(datax)
            max_value = np.max(datax)
    else:
        if min_value is None:
            min_value = np.min(datax)
        if max_value is None:
            max_value = np.max(datax)

    datax = datax[np.where(datax < max_value)]
    datax = datax[np.where(datax > min_value)]

    if abs_step is not None:
        step = abs_step
    else:
        if step is None:
            step = 5.0
        step = np.sqrt(np.median((datax - np.median(datax)) ** 2)) / step

    bins_number = int((max_value - min_value) / step) + 1
    bins = np.array(min_value + step / 2. + np.arange(bins_number) * step)

    counts = np.bincount(np.int_((datax - min_value) / step))
    counts = np.insert(counts, len(counts), np.zeros(int(bins_number) - len(counts)))

    if confidence_interval is None and not gaussian_fit:

        return bins, counts

    elif confidence_interval is not None and not gaussian_fit:

        distrx = bins
        bin_width = bins[1] - bins[0]
        distr = counts

        # corresponds to the 1-sigma level probability

        pleft = 0.0
        centroid = np.argmax(distr)
        exp_val = distrx[centroid]

        total_probability_left = np.sum(bin_width * distr[:centroid]) * confidence_interval
        total_probability_right = np.sum(bin_width * distr[centroid:]) * confidence_interval

        num = centroid
        leftci = 0
        while pleft <= total_probability_left:
            if num == centroid:
                pleft += (bin_width / 2.0) * distr[num]
            else:
                pleft += bin_width * distr[num]
            leftci = distrx[num]
            num -= 1
            if num < 0:
                print('ERROR : confidence level can not be reached from left')
                break
        pright = 0.0
        num = centroid
        rightci = 0
        while pright <= total_probability_right:
            if num == centroid:
                pright += (bin_width / 2.0) * distr[num]
            else:
                pright += bin_width * distr[num]
            rightci = distrx[num]
            num += 1
            if num > len(distr) - 1:
                print('ERROR : confidence level can not be reached from right')
                break

        value, p_error, m_error = exp_val, rightci - exp_val, exp_val - leftci

        return bins, counts, value, m_error, p_error

    elif confidence_interval is None and gaussian_fit:

        popt, pcov = fit_gaussian(bins, counts, positive=True, sampled=True)

        value, p_error, m_error = popt[2], popt[3], popt[3]

        return bins, counts, value, m_error, p_error

    else:

        raise PyLCProcessError('For the uncertainties in your distribution you should use either a confidence '
                               'interval or a gaussian fit, not both.')


def two_d_distribution(datax, datay, step=None, abs_step_x=None, abs_step_y=None, min_value_x=None, min_value_y=None,
                       max_value_x=None, max_value_y=None, mad_filter=None, confidence_interval=None,
                       gaussian_fit=False):

    datax = datax.flatten()
    datay = datay.flatten()

    if min_value_x is None and max_value_x is None:
        if mad_filter:
            data_median = np.median(datax)
            data_mad = np.sqrt(np.median((datax - data_median) ** 2))
            min_value_x = data_median - mad_filter * data_mad
            max_value_x = data_median + mad_filter * data_mad
        else:
            min_value_x = np.min(datax)
            max_value_x = np.max(datax)
    else:
        if min_value_x is None:
            min_value_x = np.min(datax)
        if max_value_x is None:
            max_value_x = np.max(datax)

    if min_value_y is None and max_value_y is None:
        if mad_filter:
            data_median = np.median(datay)
            data_mad = np.sqrt(np.median((datay - data_median) ** 2))
            min_value_y = data_median - mad_filter * data_mad
            max_value_y = data_median + mad_filter * data_mad
        else:
            min_value_y = np.min(datay)
            max_value_y = np.max(datay)
    else:
        if min_value_y is None:
            min_value_y = np.min(datay)
        if max_value_y is None:
            max_value_y = np.max(datay)

    values_to_keep = np.where((datax < max_value_x) * (datax > min_value_x) *
                              (datay < max_value_y) * (datay > min_value_y))

    datax = datax[values_to_keep]
    datay = datay[values_to_keep]

    if abs_step_x is not None:
        step_x = abs_step_x
    else:
        if step is None:
            step_x = 5.0
        else:
            step_x = step
        step_x = np.sqrt(np.median((datax - np.median(datax)) ** 2)) / step_x

    bins_number_x = int((max_value_x - min_value_x) / step_x) + 1
    bins_x = np.array(min_value_x + step_x / 2. + np.arange(bins_number_x) * step_x)

    if abs_step_y is not None:
        step_y = abs_step_y
    else:
        if step is None:
            step_y = 5.0
        else:
            step_y = step
        step_y = np.sqrt(np.median((datay - np.median(datay)) ** 2)) / step_y

    bins_number_y = int((max_value_y - min_value_y) / step_y) + 1
    bins_y = np.array(min_value_y + step_y / 2. + np.arange(bins_number_y) * step_y)

    bins_number_xy = bins_number_x * bins_number_y
    bins_xy = np.meshgrid(bins_x, bins_y)
    counts_xy = np.bincount((np.int_((datay - min_value_y) / step_y)) * bins_number_x +
                            (np.int_((datax - min_value_x) / step_x)))
    counts_xy = np.insert(counts_xy, len(counts_xy), np.zeros(bins_number_xy - len(counts_xy)))
    counts_xy = np.reshape(counts_xy, (bins_number_y, bins_number_x))

    if confidence_interval is None and not gaussian_fit:

        return bins_xy[0], bins_xy[1], counts_xy

    elif confidence_interval is not None and not gaussian_fit:

        value_x, m_error_x, p_error_x = one_d_distribution(
            datax, step=step, abs_step=abs_step_x, min_value=min_value_x, max_value=max_value_x,
            confidence_interval=confidence_interval)[2:]

        value_y, m_error_y, p_error_y = one_d_distribution(
            datay, step=step, abs_step=abs_step_y, min_value=min_value_y, max_value=max_value_y,
            confidence_interval=confidence_interval)[2:]

        return bins_xy[0], bins_xy[1], counts_xy, value_x, m_error_x, p_error_x, value_y, m_error_y, p_error_y

    elif confidence_interval is None and gaussian_fit:

        popt, pcov = fit_two_d_gaussian(bins_xy[0], bins_xy[1], counts_xy, positive=True)

        value_x, p_error_x, m_error_x = popt[2], popt[4], popt[4]
        value_y, p_error_y, m_error_y = popt[3], popt[5], popt[5]

        return bins_xy[0], bins_xy[1], counts_xy, value_x, m_error_x, p_error_x, value_y, m_error_y, p_error_y

    else:

        raise PyLCProcessError('For the uncertainties in your distribution you should use either a confidence '
                               'interval or a gaussian fit, not both.')




__all__ = ['open_dict', 'save_dict', 'copy_dict',
           'open_yaml', 'save_yaml', 'copy_yaml',
           'open_fits', 'save_fits', 'copy_fits',
           'fits_sci', 'fits_err', 'fits_sci_err',
           'download'
           ]

import os
import ssl
import copy
import yaml
import numpy as np
import pickle
import urllib

from urllib.request import urlretrieve
from astropy.io import fits as pf


def open_dict(path):
    try:
        return pickle.load(open(path, 'rb'))
    except UnicodeDecodeError:
        return pickle.load(open(path, 'rb'), encoding='latin-1')


def save_dict(dictionary, path):
    internal_copy = copy_dict(dictionary)
    pickle.dump(dictionary, open(path, 'wb'), protocol=2)
    del internal_copy


def copy_dict(dictionary):
    return copy.deepcopy(dictionary)


def open_yaml(path):
    return yaml.load(open(path, 'r'), Loader=yaml.SafeLoader)


def save_yaml(dictionary, path):
    yaml.dump(dictionary, open(path, 'w'), default_flow_style=False)


def copy_yaml(dictionary):
    return copy.deepcopy(dictionary)


def open_fits(path):
    with pf.open(path, memmap=False) as hdulist:
        internal_copy = copy_fits(hdulist)
    internal_copy.verify('fix')
    return internal_copy


def save_fits(hdulist, path):
    internal_copy = copy_fits(hdulist)
    try:
        internal_copy.writeto(path, output_verify='fix')
    except pf.VerifyError as e:
        hdu = int(str(e.args)[4:-4].split('\\n')[1].replace('HDU ', '').replace(':', ''))
        card = int(str(e.args)[4:-4].split('\\n')[2].replace('Card ', '').replace(':', ''))
        del internal_copy[hdu].header[card]
        internal_copy.writeto(path, output_verify='fix')
    
    del internal_copy


def copy_fits(hdulist):
    return pf.HDUList([ff.copy() for ff in hdulist])


def fits_sci(hdulist):
    return np.where(np.array([ff.name for ff in hdulist]) == 'SCI')[0]


def fits_err(hdulist):
    return np.where(np.array([ff.name for ff in hdulist]) == 'ERR')[0]


def fits_sci_err(hdulist):
    return np.swapaxes([fits_sci(hdulist), fits_err(hdulist)], 0, 1)


ctx = ssl.create_default_context()
ctx.check_hostname = False
ctx.verify_mode = ssl.CERT_NONE


def download(link, destination, filename=None):

    if not filename:
        filename = os.path.split(destination)[1]

    print('    Downloading {0}...'.format(filename))
    try:
        urlretrieve(link, destination)
        print('    Done!')
        return True
    except:
        try:
            with urllib.request.urlopen(link, context=ctx) as u, \
                    open(destination, 'wb') as f:
                f.write(u.read())
            print('    Done!')
            return True
        except:
            print('    Could not download {0}'.format(filename))
            return False




__all__ = ['curve_fit']

import warnings

from scipy.optimize import curve_fit as scipy_curve_fit


def curve_fit(*args, **kwargs):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message='Covariance of the parameters could not be estimated')
        return scipy_curve_fit(*args, **kwargs)




import sys
import time
import datetime


class Counter:

    def __init__(self, counter, total_iterations, show_every=1):

        self.counter = counter
        self.current_iteration = 0
        self.total_iterations = int(total_iterations)
        self.start_time = time.time()
        self.show = 0
        self.show_every = int(show_every)

    def update(self, message=''):

        self.current_iteration += 1
        self.show += 1.0 / self.show_every

        out_of = '{0}{1} / {2} {3}'.format(' ' * (len(str(self.total_iterations)) - len(str(self.current_iteration))),
                                           str(self.current_iteration), str(self.total_iterations), message)

        delta_time = time.time() - self.start_time

        time_left = str(datetime.timedelta(
            seconds=int((self.total_iterations - self.current_iteration) * delta_time / self.current_iteration)))

        total_time = str(datetime.timedelta(seconds=int(delta_time)))

        if int(self.show):

            sys.stdout.write('\r\033[K')
            sys.stdout.write('{0}{1}: {2}   time left: {3}   total time: {4}'.format(
                self.counter, '.' * (15 - len(self.counter)), out_of, time_left, total_time))
            sys.stdout.flush()

            self.show = 0

        if self.current_iteration == self.total_iterations and self.total_iterations > 1:

            print('')




__all__ = ['FixedTarget']

import numpy as np

from pylightcurve.errors import *
from pylightcurve.__databases__ import plc_data
from pylightcurve.analysis.curve_fit import curve_fit
from pylightcurve.spacetime.angles import arccos, _request_angle


class _Target:

    def __init__(self, ra, dec):

        _request_angle(ra)
        _request_angle(dec)

        if 270 > dec.deg() > 90:
            raise PyLCInputError('Declination must within -90, 90 degrees')

        self.dec = dec
        self.ra = ra

        self.coord = '{0} {1}'.format(self.ra.hms(), self.dec.dms_coord())


class FixedTarget(_Target):

    def __init__(self, ra, dec):

        _Target.__init__(self, ra, dec)

    def convert_to_bjd_tdb(self, time, time_format):

        if isinstance(time, float):

            if time_format in ['BJD_TDB', 'BJD_TT']:
                return time
            elif time_format == 'JD_UTC':
                return self._bjd_tdb(None, time)
            elif time_format == 'MJD_UTC':
                return self._bjd_tdb(None, time + 2400000.5)
            elif time_format == 'BJD_UTC':
                return self._bjd_tdb(None, curve_fit(self._bjd_utc, [0], [time], p0=[time])[0][0])
            elif time_format in ['HJD_TDB', 'HJD_TT']:
                return self._bjd_tdb(None, curve_fit(self._hjd_tdb, [0], [time], p0=[time])[0][0])
            elif time_format == 'HJD_UTC':
                return self._bjd_tdb(None, curve_fit(self._hjd_utc, [0], [time], p0=[time])[0][0])
            else:
                raise PyLCInputError(
                    'Not valid time format. Available formats: JD_UTC, MJD_UTC, BJD_UTC, BJD_TDB, BJD_TT, '
                    'HJD_UTC, HJD_BJD, HJD_TT')

        else:
            try:
                time = np.array(time, dtype=float)

                if time_format in ['BJD_TDB', 'BJD_TT']:
                    return time
                elif time_format == 'JD_UTC':
                    return np.array([self._bjd_tdb(None, ff) for ff in time])
                elif time_format == 'MJD_UTC':
                    return np.array([self._bjd_tdb(None, ff + 2400000.5) for ff in time])
                elif time_format == 'BJD_UTC':
                    return np.array([self._bjd_tdb(None, curve_fit(self._bjd_utc, [0], [ff], p0=[ff])[0][0])
                                     for ff in time])
                elif time_format in ['HJD_TDB', 'HJD_TT']:
                    return np.array([self._bjd_tdb(None, curve_fit(self._hjd_tdb, [0], [ff], p0=[ff])[0][0])
                                     for ff in time])
                elif time_format == 'HJD_UTC':
                    return np.array([self._bjd_tdb(None, curve_fit(self._hjd_utc, [0], [ff], p0=[ff])[0][0])
                                     for ff in time])
                else:
                    raise PyLCInputError(
                        'Not valid time format. Available formats: JD_UTC, MJD_UTC, BJD_UTC, BJD_TDB, BJD_TT, '
                        'HJD_UTC, HJD_BJD, HJD_TT')

            except:
                raise PyLCInputError('Not valid input for time')

    def convert_to_jd(self, time, time_format):

        if isinstance(time, float):

            if time_format in ['BJD_TDB', 'BJD_TT']:
                return curve_fit(self._bjd_tdb, [0], [time], p0=[time])[0][0]
            elif time_format == 'JD_UTC':
                return time
            elif time_format == 'MJD_UTC':
                return time + 2400000.5
            elif time_format == 'BJD_UTC':
                return curve_fit(self._bjd_utc, [0], [time], p0=[time])[0][0]
            elif time_format in ['HJD_TDB', 'HJD_TT']:
                return curve_fit(self._hjd_tdb, [0], [time], p0=[time])[0][0]
            elif time_format == 'HJD_UTC':
                return curve_fit(self._hjd_utc, [0], [time], p0=[time])[0][0]
            else:
                raise PyLCInputError(
                    'Not valid time format. Available formats: JD_UTC, MJD_UTC, BJD_UTC, BJD_TDB, BJD_TT, '
                    'HJD_UTC, HJD_BJD, HJD_TT')

        else:
            try:
                time = np.array(time, dtype=float)

                if time_format in ['BJD_TDB', 'BJD_TT']:
                    return np.array([curve_fit(self._bjd_tdb, [0], [ff], p0=[time])[0][0] for ff in time])
                elif time_format == 'JD_UTC':
                    return time
                elif time_format == 'MJD_UTC':
                    return time + 2400000.5
                elif time_format == 'BJD_UTC':
                    return np.array([curve_fit(self._bjd_utc, [0], [ff], p0=[time])[0][0] for ff in time])
                elif time_format in ['HJD_TDB', 'HJD_TT']:
                    return np.array([curve_fit(self._hjd_tdb, [0], [ff], p0=[time])[0][0] for ff in time])
                elif time_format == 'HJD_UTC':
                    return np.array([curve_fit(self._hjd_utc, [0], [ff], p0=[time])[0][0] for ff in time])
                else:
                    raise PyLCInputError(
                        'Not valid time format. Available formats: JD_UTC, MJD_UTC, BJD_UTC, BJD_TDB, BJD_TT, '
                        'HJD_UTC, HJD_BJD, HJD_TT')

            except:
                raise PyLCInputError('Not valid input for time')

    def convert_to_mjd(self, time, time_format):

        return self.convert_to_jd(time, time_format) - 2400000.5

    def _hjd_utc(self, x, jd):

        ssb_ra, ssb_dec, ssb_d, ssb_dt = plc_data.heliocentre(jd)

        a = ssb_d / 60.0 / 24.0
        b = self.dec.sin() * np.sin(ssb_dec)
        c = self.dec.cos() * np.cos(ssb_dec) * np.cos(self.ra.rad() - ssb_ra)

        return jd - a * (b + c)

    def _hjd_tdb(self, x, jd):

        ssb_ra, ssb_dec, ssb_d, ssb_dt = plc_data.heliocentre(jd)

        a = ssb_d / 60.0 / 24.0
        b = self.dec.sin() * np.sin(ssb_dec)
        c = self.dec.cos() * np.cos(ssb_dec) * np.cos(self.ra.rad() - ssb_ra)

        return jd - a * (b + c) + ssb_dt / 60.0 / 60.0 / 24.0

    def _bjd_utc(self, x, jd):

        ssb_ra, ssb_dec, ssb_d, ssb_dt = plc_data.barycentre(jd)

        a = ssb_d / 60.0 / 24.0
        b = self.dec.sin() * np.sin(ssb_dec)
        c = self.dec.cos() * np.cos(ssb_dec) * np.cos(self.ra.rad() - ssb_ra)

        return jd - a * (b + c)

    def _bjd_tdb(self, x, jd):

        ssb_ra, ssb_dec, ssb_d, ssb_dt = plc_data.barycentre(jd)

        a = ssb_d / 60.0 / 24.0
        b = self.dec.sin() * np.sin(ssb_dec)
        c = self.dec.cos() * np.cos(ssb_dec) * np.cos(self.ra.rad() - ssb_ra)

        return jd - a * (b + c) + ssb_dt / 60.0 / 60.0 / 24.0

    def distance_on_sphere(self, other):

        _request_target(other)

        return arccos(self.dec.sin() * other.dec.sin() + self.dec.cos() * other.dec.cos() * (self.ra - other.ra).cos())

    def __str__(self):
        return 'plc.FixedTarget(RA(hms)/DEC(dms): {0})'.format(self.coord)

    def __repr__(self):
        return self.__str__()


def _is_target(item):
    if isinstance(item, FixedTarget):
        return True
    else:
        return False


def _request_target(item):
    if _is_target(item):
        pass
    else:
        raise PyLCInputError('A plc.Target object is required (plc.FixedTarget)')
